"""
Definition for the ObjectInfo NGA modified RDE/AFRL labeling object
"""

__classification__ = "UNCLASSIFIED"
__authors__ = "Thomas McCullough"

import logging
from typing import Optional, List

import numpy

from sarpy.io.xml.base import Serializable, Arrayable, create_text_node, \
    get_node_value
from sarpy.io.xml.descriptors import StringDescriptor, FloatDescriptor, \
    IntegerDescriptor, SerializableDescriptor, SerializableListDescriptor
from sarpy.io.complex.sicd_elements.blocks import RowColType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.product.sidd2_elements.SIDD import SIDDType
from sarpy.geometry.geocoords import geodetic_to_ecf, ecf_to_geodetic, wgs_84_norm
from sarpy.geometry.geometry_elements import Point, Polygon, GeometryCollection, Geometry
from sarpy.annotation.base import GeometryProperties

from .base import DEFAULT_STRICT
from .blocks import RangeCrossRangeType, RowColDoubleType, LatLonEleType, \
    ProjectionPerturbationType, LabelSourceType


logger = logging.getLogger(__name__)

_no_projection_text = 'This sicd does not permit projection,\n\t' \
                      'so the image location can not be inferred'


# the Object and sub-component definitions

class PhysicalType(Serializable):
    _fields = ('ChipSize', 'CenterPixel')
    _required = _fields
    ChipSize = SerializableDescriptor(
        'ChipSize', RangeCrossRangeType, _required, strict=DEFAULT_STRICT,
        docstring='The chip size of the physical object, '
                  'in the appropriate plane')  # type: RangeCrossRangeType
    CenterPixel = SerializableDescriptor(
        'CenterPixel', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='The center pixel of the physical object, '
                  'in the appropriate plane')  # type: RowColDoubleType

    def __init__(self, ChipSize=None, CenterPixel=None, **kwargs):
        """
        Parameters
        ----------
        ChipSize : RangeCrossRangeType|numpy.ndarray|list|tuple
        CenterPixel : RowColDoubleType|numpy.ndarray|list|tuple
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ChipSize = ChipSize
        self.CenterPixel = CenterPixel
        super(PhysicalType, self).__init__(**kwargs)

    @classmethod
    def from_ranges(cls, row_range, col_range, row_limit, col_limit):
        """
        Construct from the row/column ranges and limits.

        Parameters
        ----------
        row_range
        col_range
        row_limit
        col_limit

        Returns
        -------
        PhysicalType
        """

        first_row, last_row = max(0, row_range[0]), min(row_limit, row_range[1])
        first_col, last_col = max(0, col_range[0]), min(col_limit, col_range[1])
        return PhysicalType(
            ChipSize=(last_row-first_row, last_col-first_col),
            CenterPixel=(0.5*(last_row+first_row), 0.5*(last_col+first_col)))


class PlanePhysicalType(Serializable):
    _fields = (
        'Physical', 'PhysicalWithShadows')
    _required = _fields
    Physical = SerializableDescriptor(
        'Physical', PhysicalType, _required,
        docstring='Chip details for the physical object in the appropriate plane')  # type: PhysicalType
    PhysicalWithShadows = SerializableDescriptor(
        'PhysicalWithShadows', PhysicalType, _required,
        docstring='Chip details for the physical object including shadows in '
                  'the appropriate plane')  # type: PhysicalType

    def __init__(self, Physical=None, PhysicalWithShadows=None, **kwargs):
        """

        Parameters
        ----------
        Physical : PhysicalType
        PhysicalWithShadows : PhysicalType
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Physical = Physical
        self.PhysicalWithShadows = PhysicalWithShadows
        super(PlanePhysicalType, self).__init__(**kwargs)


class SizeType(Serializable, Arrayable):
    _fields = ('Length', 'Width', 'Height')
    _required = _fields
    _numeric_format = {key: '0.17G' for key in _fields}
    # Descriptors
    Length = FloatDescriptor(
        'Length', _required, strict=True, docstring='The Length attribute.')  # type: float
    Width = FloatDescriptor(
        'Width', _required, strict=True, docstring='The Width attribute.')  # type: float
    Height = FloatDescriptor(
        'Height', _required, strict=True, docstring='The Height attribute.')  # type: float

    def __init__(self, Length=None, Width=None, Height=None, **kwargs):
        """
        Parameters
        ----------
        Length : float
        Width : float
        Height : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Length, self.Width, self.Height = Length, Width, Height
        super(SizeType, self).__init__(**kwargs)

    def get_max_diameter(self):
        """
        Gets the nominal maximum diameter for the item, in meters.

        Returns
        -------
        float
        """

        return float(numpy.sqrt(self.Length*self.Length + self.Width*self.Width))

    def get_array(self, dtype='float64'):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Length, Width, Height]
        """

        return numpy.array([self.Length, self.Width, self.Height], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Length, Width, Height]

        Returns
        -------
        SizeType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 3:
                raise ValueError('Expected array to be of length 3, and received {}'.format(array))
            return cls(Length=array[0], Width=array[1], Height=array[2])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class OrientationType(Serializable):
    _fields = ('Roll', 'Pitch', 'Yaw', 'AzimuthAngle')
    _required = ()
    _numeric_format = {key: '0.17G' for key in _fields}
    # descriptors
    Roll = FloatDescriptor(
        'Roll', _required)  # type: float
    Pitch = FloatDescriptor(
        'Pitch', _required)  # type: float
    Yaw = FloatDescriptor(
        'Yaw', _required)  # type: float
    AzimuthAngle = FloatDescriptor(
        'AzimuthAngle', _required)  # type: float

    def __init__(self, Roll=None, Pitch=None, Yaw=None, AzimuthAngle=None, **kwargs):
        """
        Parameters
        ----------
        Roll : float
        Pitch : float
        Yaw : float
        AzimuthAngle : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Roll = Roll
        self.Pitch = Pitch
        self.Yaw = Yaw
        self.AzimuthAngle = AzimuthAngle
        super(OrientationType, self).__init__(**kwargs)


class ImageLocationType(Serializable):
    _fields = (
        'CenterPixel', 'LeftFrontPixel', 'RightFrontPixel', 'RightRearPixel',
        'LeftRearPixel')
    _required = ('CenterPixel', )
    # descriptors
    CenterPixel = SerializableDescriptor(
        'CenterPixel', RowColType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: RowColType
    LeftFrontPixel = SerializableDescriptor(
        'LeftFrontPixel', RowColType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[RowColType]
    RightFrontPixel = SerializableDescriptor(
        'RightFrontPixel', RowColType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[RowColType]
    RightRearPixel = SerializableDescriptor(
        'RightRearPixel', RowColType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[RowColType]
    LeftRearPixel = SerializableDescriptor(
        'LeftRearPixel', RowColType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[RowColType]

    def __init__(self, CenterPixel=None, LeftFrontPixel=None, RightFrontPixel=None,
                 RightRearPixel=None, LeftRearPixel=None, **kwargs):
        """
        Parameters
        ----------
        CenterPixel : RowColType|numpy.ndarray|list|tuple
        LeftFrontPixel : RowColType|numpy.ndarray|list|tuple
        RightFrontPixel : RowColType|numpy.ndarray|list|tuple
        RightRearPixel : RowColType|numpy.ndarray|list|tuple
        LeftRearPixel : RowColType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CenterPixel = CenterPixel
        self.LeftFrontPixel = LeftFrontPixel
        self.RightFrontPixel = RightFrontPixel
        self.RightRearPixel = RightRearPixel
        self.LeftRearPixel = LeftRearPixel
        super(ImageLocationType, self).__init__(**kwargs)

    @classmethod
    def from_geolocation(cls, geo_location, the_structure):
        """
        Construct the image location from the geographical location via
        projection using the SICD model.

        Parameters
        ----------
        geo_location : GeoLocationType
        the_structure : SICDType|SIDDType

        Returns
        -------
        None|ImageLocationType
            None if projection fails, the value otherwise
        """

        if geo_location is None:
            return None

        if not the_structure.can_project_coordinates():
            logger.warning(_no_projection_text)
            return None

        # make sure this is defined, for the sake of efficiency
        the_structure.define_coa_projection(override=False)

        kwargs = {}

        if isinstance(the_structure, SICDType):
            image_shift = numpy.array(
                [the_structure.ImageData.FirstRow, the_structure.ImageData.FirstCol], dtype='float64')
        else:
            image_shift = numpy.zeros((2, ), dtype='float64')

        for attribute in cls._fields:
            value = getattr(geo_location, attribute)
            if value is not None:
                absolute_pixel_location, _, _ = the_structure.project_ground_to_image_geo(
                    value.get_array(dtype='float64'), ordering='latlong')
                if numpy.any(numpy.isnan(absolute_pixel_location)):
                    return None

                kwargs[attribute] = absolute_pixel_location - image_shift
        out = ImageLocationType(**kwargs)
        out.infer_center_pixel()
        return out

    def infer_center_pixel(self):
        """
        Infer the center pixel, if not populated.

        Returns
        -------
        None
        """

        if self.CenterPixel is not None:
            return

        current = numpy.zeros((2, ), dtype='float64')
        for entry in self._fields:
            if entry == 'CenterPixel':
                continue
            value = getattr(self, entry)
            if value is None:
                return
            current += 0.25*value.get_array(dtype='float64')
        self.CenterPixel = RowColType.from_array(current)

    def get_nominal_box(self, row_length=10, col_length=10):
        """
        Get a nominal box containing the object, using the default side length if necessary.

        Parameters
        ----------
        row_length : int|float
            The side length to use for the rectangle, if not defined.
        col_length : int|float
            The side length to use for the rectangle, if not defined.

        Returns
        -------
        None|numpy.ndarray
        """

        if self.LeftFrontPixel is not None and self.RightFrontPixel is not None and \
                self.LeftRearPixel is not None and self.RightRearPixel is not None:
            out = numpy.zeros((4, 2), dtype='float64')
            out[0, :] = self.LeftFrontPixel.get_array()
            out[1, :] = self.RightFrontPixel.get_array()
            out[2, :] = self.RightRearPixel.get_array()
            out[3, :] = self.LeftRearPixel.get_array()
            return out

        if self.CenterPixel is None:
            return None

        shift = numpy.array([[-0.5, -0.5], [-0.5, 0.5], [0.5, 0.5], [0.5, -0.5]], dtype='float64')
        shift[:, 0] *= row_length
        shift[:, 1] *= col_length
        return self.CenterPixel.get_array(dtype='float64') + shift

    def get_geometry_object(self):
        """
        Gets the geometry for the given image section.

        Returns
        -------
        geometry : None|Point|GeometryCollection
        geometry_properties : None|List[GeometryProperties]
        """

        geometries = []
        geometry_properties = []
        if self.CenterPixel is not None:
            geometries.append(Point(coordinates=self.CenterPixel.get_array(dtype='float64')))
            geometry_properties.append(GeometryProperties(name='CenterPixel', color='blue'))
        if self.LeftFrontPixel is not None and \
                self.RightFrontPixel is not None and \
                self.RightRearPixel is not None and \
                self.LeftRearPixel is not None:
            ring = numpy.zeros((4, 2), dtype='float64')
            ring[0, :] = self.LeftFrontPixel.get_array(dtype='float64')
            ring[1, :] = self.RightFrontPixel.get_array(dtype='float64')
            ring[2, :] = self.RightRearPixel.get_array(dtype='float64')
            ring[3, :] = self.LeftRearPixel.get_array(dtype='float64')
            geometries.append(Polygon(coordinates=[ring, ]))
            geometry_properties.append(GeometryProperties(name='Polygon', color='green'))

        if len(geometries) == 0:
            return None, None
        elif len(geometries) == 1:
            return geometries[0], geometry_properties
        else:
            return GeometryCollection(geometries=geometries), geometry_properties


class GeoLocationType(Serializable):
    _fields = (
        'CenterPixel', 'LeftFrontPixel', 'RightFrontPixel', 'RightRearPixel',
        'LeftRearPixel')
    _required = ('CenterPixel', )
    # descriptors
    CenterPixel = SerializableDescriptor(
        'CenterPixel', LatLonEleType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: LatLonEleType
    LeftFrontPixel = SerializableDescriptor(
        'LeftFrontPixel', LatLonEleType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[LatLonEleType]
    RightFrontPixel = SerializableDescriptor(
        'RightFrontPixel', LatLonEleType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[LatLonEleType]
    RightRearPixel = SerializableDescriptor(
        'RightRearPixel', LatLonEleType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[LatLonEleType]
    LeftRearPixel = SerializableDescriptor(
        'LeftRearPixel', LatLonEleType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[LatLonEleType]

    def __init__(self, CenterPixel=None, LeftFrontPixel=None, RightFrontPixel=None,
                 RightRearPixel=None, LeftRearPixel=None, **kwargs):
        """
        Parameters
        ----------
        CenterPixel : LatLonEleType|numpy.ndarray|list|tuple
        LeftFrontPixel : None|LatLonEleType|numpy.ndarray|list|tuple
        RightFrontPixel : None|LatLonEleType|numpy.ndarray|list|tuple
        RightRearPixel : None|LatLonEleType|numpy.ndarray|list|tuple
        LeftRearPixel : None|LatLonEleType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CenterPixel = CenterPixel
        self.LeftFrontPixel = LeftFrontPixel
        self.RightFrontPixel = RightFrontPixel
        self.RightRearPixel = RightRearPixel
        self.LeftRearPixel = LeftRearPixel
        super(GeoLocationType, self).__init__(**kwargs)

    # noinspection PyUnusedLocal
    @classmethod
    def from_image_location(cls, image_location, the_structure, projection_type='HAE', **proj_kwargs):
        """
        Construct the geographical location from the image location via
        projection using the SICD model.

        .. Note::
            This assumes that the image coordinates are with respect to the given
            image (chip), and NOT including any sicd.ImageData.FirstRow/Col values,
            which will be added here.

        Parameters
        ----------
        image_location : ImageLocationType
        the_structure : SICDType|SIDDType
        projection_type : str
            The projection type selector, one of `['PLANE', 'HAE', 'DEM']`. Using `'DEM'`
            requires configuration for the DEM pathway described in
            :func:`sarpy.geometry.point_projection.image_to_ground_dem`.
        proj_kwargs
            The keyword arguments for the :func:`SICDType.project_image_to_ground_geo` method.

        Returns
        -------
        None|GeoLocationType
            Coordinates may be populated as `NaN` if projection fails.
        """

        if image_location is None:
            return None

        if not the_structure.can_project_coordinates():
            logger.warning(_no_projection_text)
            return None

        # make sure this is defined, for the sake of efficiency
        the_structure.define_coa_projection(override=False)

        if isinstance(the_structure, SICDType):
            image_shift = numpy.array(
                [the_structure.ImageData.FirstRow, the_structure.ImageData.FirstCol], dtype='float64')
        else:
            image_shift = numpy.zeros((2, ), dtype='float64')

        kwargs = {}
        for attribute in cls._fields:
            value = getattr(image_location, attribute)
            if value is not None:
                coords = value.get_array(dtype='float64') + image_shift
                geo_coords = the_structure.project_image_to_ground_geo(
                    coords, ordering='latlong', projection_type=projection_type, **proj_kwargs)

                kwargs[attribute] = geo_coords
        out = GeoLocationType(**kwargs)
        out.infer_center_pixel()
        return out

    def infer_center_pixel(self):
        """
        Infer the center pixel, if not populated.

        Returns
        -------
        None
        """

        if self.CenterPixel is not None:
            return

        current = numpy.zeros((3, ), dtype='float64')
        for entry in self._fields:
            if entry == 'CenterPixel':
                continue
            value = getattr(self, entry)
            if value is None:
                return
            current += 0.25*geodetic_to_ecf(value.get_array(dtype='float64'))
        self.CenterPixel = LatLonEleType.from_array(ecf_to_geodetic(current))


class StringWithComponentType(Serializable):
    _fields = ('Component', 'Value')
    _required = ('Value', )
    Component = StringDescriptor(
        'Component', _required)  # type: str
    Value = StringDescriptor(
        'Value', _required)  # type: str
    _set_as_attribute = ('Component', )

    def __init__(self, Component=None, Value=None, **kwargs):
        """

        Parameters
        ----------
        Component : str
        Value : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Component = Component
        self.Value = Value
        super(StringWithComponentType, self).__init__(**kwargs)

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        value = get_node_value(node)
        component = node.attrib.get('Component', None)
        return cls(Value=value, Component=component)
    
    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):        
        if (ns_key is not None and ns_key != 'default') and not tag.startswith(ns_key + ':'):
            use_tag = '{}:{}'.format(ns_key, tag)
        else:
            use_tag = tag
        if self.Value is None:
            value = ''
        else:
            value = self.Value
        node = create_text_node(doc, use_tag, value, parent=parent)
        if self.Component is not None:
            node.attrib['Component'] = self.Component
        return node


class TheObjectType(Serializable):
    _fields = (
        'SystemName', 'SystemComponent', 'NATOName', 'Function', 'Version', 'DecoyType', 'SerialNumber',
        'ObjectClass', 'ObjectSubClass', 'ObjectTypeClass', 'ObjectType', 'ObjectLabel',
        'SlantPlane', 'GroundPlane', 'Size', 'Orientation',
        'Articulation', 'Configuration',
        'Accessories', 'PaintScheme', 'Camouflage', 'Obscuration', 'ObscurationPercent', 'ImageLevelObscuration',
        'ImageLocation', 'GeoLocation',
        'TargetToClutterRatio', 'VisualQualityMetric',
        'UnderlyingTerrain', 'OverlyingTerrain', 'TerrainTexture', 'SeasonalCover', 
        'ProjectionPerturbation')
    _required = ('SystemName', 'ImageLocation', 'GeoLocation')
    _numeric_format = {'ObscurationPercent': '0.17G', }
    _collections_tags = {
        'Articulation': {'array': False, 'child_tag': 'Articulation'}, 
        'Configuration': {'array': False, 'child_tag': 'Configuration'}}
    # descriptors
    SystemName = StringDescriptor(
        'SystemName', _required, strict=DEFAULT_STRICT,
        docstring='Name of the object.')  # type: str
    SystemComponent = StringDescriptor(
        'SystemComponent', _required, strict=DEFAULT_STRICT,
        docstring='Name of the weapon system component.')  # type: Optional[str]
    NATOName = StringDescriptor(
        'NATOName', _required, strict=DEFAULT_STRICT,
        docstring='Name of the object in NATO naming convention.')  # type: Optional[str]
    Function = StringDescriptor(
        'Function', _required, strict=DEFAULT_STRICT,
        docstring='Function of the object.')  # type: Optional[str]
    Version = StringDescriptor(
        'Version', _required, strict=DEFAULT_STRICT,
        docstring='Version number of the object.')  # type: Optional[str]
    DecoyType = StringDescriptor(
        'DecoyType', _required, strict=DEFAULT_STRICT,
        docstring='Object is a decoy or surrogate.')  # type: Optional[str]
    SerialNumber = StringDescriptor(
        'SerialNumber', _required, strict=DEFAULT_STRICT,
        docstring='Serial number of the object.')  # type: Optional[str]
    # label elements
    ObjectClass = StringDescriptor(
        'ObjectClass', _required, strict=DEFAULT_STRICT,
        docstring='Top level class indicator; e.g., Aircraft, Ship, '
                  'Ground Vehicle, Missile Launcher, etc.')  # type: Optional[str]
    ObjectSubClass = StringDescriptor(
        'ObjectSubClass', _required, strict=DEFAULT_STRICT,
        docstring='Sub-class indicator; e.g., military, commercial')  # type: Optional[str]
    ObjectTypeClass = StringDescriptor(
        'ObjectTypeClass', _required, strict=DEFAULT_STRICT,
        docstring='Object type class indicator; e.g., '
                  'for Aircraft/Military - Propeller, Jet')  # type: Optional[str]
    ObjectType = StringDescriptor(
        'ObjectType', _required, strict=DEFAULT_STRICT,
        docstring='Object type indicator, e.g., '
                  'for Aircraft/Military/Jet - Bomber, Fighter')  # type: Optional[str]
    ObjectLabel = StringDescriptor(
        'ObjectLabel', _required, strict=DEFAULT_STRICT,
        docstring='Object label indicator, e.g., '
                  'for Bomber - Il-28, Tu-22M, Tu-160')  # type: Optional[str]
    SlantPlane = SerializableDescriptor(
        'SlantPlane', PlanePhysicalType, _required, strict=DEFAULT_STRICT,
        docstring='Object physical definition in the slant plane')  # type: Optional[PlanePhysicalType]
    GroundPlane = SerializableDescriptor(
        'GroundPlane', PlanePhysicalType, _required, strict=DEFAULT_STRICT,
        docstring='Object physical definition in the ground plane')  # type: Optional[PlanePhysicalType]
    # specific physical quantities
    Size = SerializableDescriptor(
        'Size', SizeType, _required, strict=DEFAULT_STRICT,
        docstring='The actual physical size of the object')  # type: Optional[SizeType]
    Orientation = SerializableDescriptor(
        'Orientation', OrientationType, _required, strict=DEFAULT_STRICT,
        docstring='The actual orientation size of the object')  # type: Optional[OrientationType]
    Articulation = SerializableListDescriptor(
        'Articulation', StringWithComponentType, _collections_tags, _required, 
        docstring='')  # type: List[StringWithComponentType]
    Configuration = SerializableListDescriptor(
        'Configuration', StringWithComponentType, _collections_tags, _required, 
        docstring='')  # type: List[StringWithComponentType]        
    Accessories = StringDescriptor(
        'Accessories', _required, strict=DEFAULT_STRICT,
        docstring='Defines items that are out of the norm, or have been added or removed.')  # type: Optional[str]
    PaintScheme = StringDescriptor(
        'PaintScheme', _required, strict=DEFAULT_STRICT,
        docstring='Paint scheme of object (e.g. olive drab, compass ghost grey, etc.).')  # type: Optional[str]
    Camouflage = StringDescriptor(
        'Camouflage', _required, strict=DEFAULT_STRICT,
        docstring='Details the camouflage on the object.')  # type: Optional[str]
    Obscuration = StringDescriptor(
        'Obscuration', _required, strict=DEFAULT_STRICT,
        docstring='General description of the obscuration.')  # type: Optional[str]
    ObscurationPercent = FloatDescriptor(
        'ObscurationPercent', _required, strict=DEFAULT_STRICT,
        docstring='The percent obscuration.')  # type: Optional[float]
    ImageLevelObscuration = StringDescriptor(
        'ImageLevelObscuration', _required, strict=DEFAULT_STRICT,
        docstring='Specific description of the obscuration based on the sensor look angle.')  # type: Optional[str]
    # location of the labeled item
    ImageLocation = SerializableDescriptor(
        'ImageLocation', ImageLocationType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: ImageLocationType
    GeoLocation = SerializableDescriptor(
        'GeoLocation', GeoLocationType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: GeoLocationType
    # text quality descriptions
    TargetToClutterRatio = StringDescriptor(
        'TargetToClutterRatio', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[str]
    VisualQualityMetric = StringDescriptor(
        'VisualQualityMetric', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[str]
    UnderlyingTerrain = StringDescriptor(
        'UnderlyingTerrain', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[str]
    OverlyingTerrain = StringDescriptor(
        'OverlyingTerrain', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[str]
    TerrainTexture = StringDescriptor(
        'TerrainTexture', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[str]
    SeasonalCover = StringDescriptor(
        'SeasonalCover', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[str]
    ProjectionPerturbation = SerializableDescriptor(
        'ProjectionPerturbation', ProjectionPerturbationType, _required, 
        docstring='')  # type: Optional[ProjectionPerturbationType]

    def __init__(self, SystemName=None, SystemComponent=None, NATOName=None,
                 Function=None, Version=None, DecoyType=None, SerialNumber=None,
                 ObjectClass=None, ObjectSubClass=None, ObjectTypeClass=None,
                 ObjectType=None, ObjectLabel=None,
                 SlantPlane=None, GroundPlane=None,
                 Size=None, Orientation=None,
                 Articulation=None, Configuration=None,
                 Accessories=None, PaintScheme=None, Camouflage=None,
                 Obscuration=None, ObscurationPercent=None, ImageLevelObscuration=None,
                 ImageLocation=None, GeoLocation=None,
                 TargetToClutterRatio=None, VisualQualityMetric=None,
                 UnderlyingTerrain=None, OverlyingTerrain=None,
                 TerrainTexture=None, SeasonalCover=None, 
                 ProjectionPerturbation=None,
                 **kwargs):
        """
        Parameters
        ----------
        SystemName : str
        SystemComponent : None|str
        NATOName : None|str
        Function : None|str
        Version : None|str
        DecoyType : None|str
        SerialNumber : None|str
        ObjectClass : None|str
        ObjectSubClass : None|str
        ObjectTypeClass : None|str
        ObjectType : None|str
        ObjectLabel : None|str
        SlantPlane : None|PlanePhysicalType
        GroundPlane : None|PlanePhysicalType
        Size : None|SizeType|numpy.ndarray|list|tuple
        Orientation : OrientationType
        Articulation : None|str|List[StringWithComponentType]
        Configuration : None|str|List[StringWithComponentType]
        Accessories : None|str
        PaintScheme : None|str
        Camouflage : None|str
        Obscuration : None|str
        ObscurationPercent : None|float
        ImageLevelObscuration : None|str
        ImageLocation : ImageLocationType
        GeoLocation : GeoLocationType
        TargetToClutterRatio : None|str
        VisualQualityMetric : None|str
        UnderlyingTerrain : None|str
        OverlyingTerrain : None|str
        TerrainTexture : None|str
        SeasonalCover : None|str
        ProjectionPerturbation : None|ProjectionPerturbationType
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SystemName = SystemName
        self.SystemComponent = SystemComponent
        self.NATOName = NATOName
        self.Function = Function
        self.Version = Version
        self.DecoyType = DecoyType
        self.SerialNumber = SerialNumber

        self.ObjectClass = ObjectClass
        self.ObjectSubClass = ObjectSubClass
        self.ObjectTypeClass = ObjectTypeClass
        self.ObjectType = ObjectType
        self.ObjectLabel = ObjectLabel

        self.SlantPlane = SlantPlane
        self.GroundPlane = GroundPlane
        self.Size = Size
        self.Orientation = Orientation

        if isinstance(Articulation, (str, dict)):
            self.add_articulation(Articulation)
        elif isinstance(Articulation, list):
            for entry in Articulation:
                self.add_articulation(entry)
        else:
            self.Articulation = Articulation

        if isinstance(Configuration, (str, dict)):
            self.add_configuration(Configuration)
        elif isinstance(Configuration, list):
            for entry in Configuration:
                self.add_configuration(entry)
        else:
            self.Configuration = Configuration

        self.Accessories = Accessories
        self.PaintScheme = PaintScheme
        self.Camouflage = Camouflage
        self.Obscuration = Obscuration
        self.ObscurationPercent = ObscurationPercent
        self.ImageLevelObscuration = ImageLevelObscuration

        self.ImageLocation = ImageLocation
        self.GeoLocation = GeoLocation

        self.TargetToClutterRatio = TargetToClutterRatio
        self.VisualQualityMetric = VisualQualityMetric
        self.UnderlyingTerrain = UnderlyingTerrain
        self.OverlyingTerrain = OverlyingTerrain
        self.TerrainTexture = TerrainTexture
        self.SeasonalCover = SeasonalCover
        self.ProjectionPerturbation = ProjectionPerturbation
        super(TheObjectType, self).__init__(**kwargs)

    def _check_placement(self, rows, cols, row_bounds, col_bounds, overlap_cutoff=0.5):
        """
        Checks the bounds condition for the provided box.

        Here inclusion is defined by what proportion of the area of the proposed
        chip is actually contained inside the image bounds.

        Parameters
        ----------
        rows : int|float
            The number of rows in the image.
        cols : int|float
            The number of columns in the image.
        row_bounds : List
            Of the form `[row min, row max]`
        col_bounds : List
            Of the form `[col min, col max]`
        overlap_cutoff : float
            Determines the transition from in the periphery to out of the image.

        Returns
        -------
        int
            1 - completely in the image
            2 - the proposed chip has `overlap_cutoff <= fractional contained area < 1`
            3 - the proposed chip has `fractional contained area < overlap_cutoff`
        """

        if row_bounds[1] <= row_bounds[0] or col_bounds[1] <= col_bounds[0]:
            raise ValueError('bounds out of order ({}, {})'.format(row_bounds, col_bounds))
        if 0 <= row_bounds[0] and rows < row_bounds[1] and 0 <= col_bounds[0] and cols < col_bounds[1]:
            return 1  # completely in bounds

        row_size = row_bounds[1] - row_bounds[0]
        col_size = col_bounds[1] - col_bounds[0]

        first_row, last_row = max(0, row_bounds[0]), min(rows, row_bounds[1])
        first_col, last_col = max(0, col_bounds[0]), min(cols, col_bounds[1])

        area_overlap = (last_row - first_row)*(last_col - first_col)
        if area_overlap >= overlap_cutoff*row_size*col_size:
            return 2  # the item is at the periphery
        else:
            return 3  # it should be considered out of range

    def set_image_location_from_sicd(self, sicd, populate_in_periphery=False):
        """
        Set the image location information with respect to the given SICD,
        assuming that the physical coordinates are populated.

        Parameters
        ----------
        sicd : SICDType
        populate_in_periphery : bool

        Returns
        -------
        int
            -1 - insufficient metadata to proceed or other failure
            0 - nothing to be done
            1 - successful
            2 - object in the image periphery, populating based on `populate_in_periphery`
            3 - object not in the image field
        """

        if self.ImageLocation is not None:
            # no need to infer anything, it's already populated
            return 0

        if self.GeoLocation is None:
            logger.warning(
                'GeoLocation is not populated,\n\t'
                'so the image location can not be inferred')
            return -1

        if not sicd.can_project_coordinates():
            logger.warning(_no_projection_text)
            return -1

        # gets the prospective image location
        image_location = ImageLocationType.from_geolocation(self.GeoLocation, sicd)
        if image_location is None:
            return -1

        self.ImageLocation = image_location
        # get nominal object size in meters and pixels
        if self.Size is None:
            row_size = 2.0
            col_size = 2.0
        else:
            max_size = self.Size.get_max_diameter()
            if max_size == 0:
                max_size = 10.0
            row_size = max_size/sicd.Grid.Row.SS
            col_size = max_size/sicd.Grid.Col.SS
        # check bounding information
        rows = sicd.ImageData.NumRows
        cols = sicd.ImageData.NumCols
        center_pixel = image_location.CenterPixel.get_array(dtype='float64')
        row_bounds = [center_pixel[0] - 0.5*row_size, center_pixel[0] + 0.5*row_size]
        col_bounds = [center_pixel[1] - 0.5*col_size, center_pixel[1] + 0.5*col_size]

        placement = self._check_placement(rows, cols, row_bounds, col_bounds)

        if placement == 3:
            return placement
        if placement == 2 and not populate_in_periphery:
            return placement

        self.ImageLocation = image_location
        return placement

    def set_geo_location_from_sicd(self, sicd, projection_type='HAE', **proj_kwargs):
        """
        Set the geographical location information with respect to the given SICD,
        assuming that the image coordinates are populated.

        .. Note::
            This assumes that the image coordinates are with respect to the given
            image (chip), and NOT including any sicd.ImageData.FirstRow/Col values,
            which will be added here.

        Parameters
        ----------
        sicd : SICDType
        projection_type : str
            The projection type selector, one of `['PLANE', 'HAE', 'DEM']`. Using `'DEM'`
            requires configuration for the DEM pathway described in
            :func:`sarpy.geometry.point_projection.image_to_ground_dem`.
        proj_kwargs
            The keyword arguments for the :func:`SICDType.project_image_to_ground_geo` method.
        """

        if self.GeoLocation is not None:
            # no need to infer anything, it's already populated
            return

        if self.ImageLocation is None:
            logger.warning(
                'ImageLocation is not populated,\n\t'
                'so the geographical location can not be inferred')
            return

        if not sicd.can_project_coordinates():
            logger.warning(_no_projection_text)
            return

        self.GeoLocation = GeoLocationType.from_image_location(
            self.ImageLocation, sicd, projection_type=projection_type, **proj_kwargs)

    def set_chip_details_from_sicd(self, sicd, layover_shift=False, populate_in_periphery=False, padding_fraction=0.05, minimum_pad=0):
        """
        Set the chip information with respect to the given SICD, assuming that the
        image location and size are defined.

        Parameters
        ----------
        sicd : SICDType
        layover_shift : bool
            Shift based on layover direction? This should be `True` if the identification of
            the bounds and/or center pixel do not include any layover, as in
            populating location from known ground truth. This should be `False` if
            the identification of bounds and/or center pixel do include layover,
            potentially as based on annotation of the imagery itself in pixel
            space.
        populate_in_periphery : bool
            Should we populate for peripheral?
        padding_fraction : None|float
            Default fraction of box dimension by which to pad.
        minimum_pad : int|float
            The minimum number of pixels by which to pad for the chip definition

        Returns
        -------
        int
            -1 - insufficient metadata to proceed
            0 - nothing to be done
            1 - successful
            2 - object in the image periphery, populating based on `populate_in_periphery`
            3 - object not in the image field
        """

        if self.SlantPlane is not None:
            # no need to infer anything, it's already populated
            return 0

        if self.Size is None:
            logger.warning(
                'Size is not populated,\n\t'
                'so the chip size can not be inferred')
            return -1

        if self.ImageLocation is None:
            # try to set from geolocation
            return_value = self.set_image_location_from_sicd(sicd, populate_in_periphery=populate_in_periphery)
            if return_value in [-1, 3] or (return_value == 2 and not populate_in_periphery):
                return return_value

        # get nominal object size, in meters
        max_size = self.Size.get_max_diameter()  # in meters
        row_size = max_size/sicd.Grid.Row.SS  # in pixels
        col_size = max_size/sicd.Grid.Col.SS  # in pixels

        # get nominal image box
        image_location = self.ImageLocation
        pixel_box = image_location.get_nominal_box(row_length=row_size, col_length=col_size)

        ground_unit_norm = wgs_84_norm(sicd.GeoData.SCP.ECF.get_array())
        slant_plane_unit_norm = numpy.cross(sicd.Grid.Row.UVectECF.get_array(), sicd.Grid.Col.UVectECF.get_array())
        magnitude_factor = ground_unit_norm.dot(slant_plane_unit_norm)
        # determines the relative size of things in slant plane versus ground plane

        # get nominal layover vector - should be pointed generally towards the top (negative rows value)
        layover_magnitude = sicd.SCPCOA.LayoverMagnitude
        if layover_magnitude is None:
            layover_magnitude = 0.25
        layover_size = self.Size.Height*layover_magnitude*magnitude_factor
        if sicd.SCPCOA.LayoverAng is None:
            layover_angle = 0.0
        else:
            layover_angle = numpy.deg2rad(sicd.SCPCOA.LayoverAng - sicd.SCPCOA.AzimAng)
        layover_vector = -layover_size*numpy.array(
            [numpy.cos(layover_angle)/sicd.Grid.Row.SS, -numpy.sin(layover_angle)/sicd.Grid.Col.SS])

        # craft the layover box
        if layover_shift:
            layover_box = pixel_box + layover_vector
        else:
            layover_box = pixel_box

        # determine the maximum and minimum pixel values here
        min_rows = min(numpy.min(pixel_box[:, 0]), numpy.min(layover_box[:, 0]))
        max_rows = max(numpy.max(pixel_box[:, 0]), numpy.max(layover_box[:, 0]))
        min_cols = min(numpy.min(pixel_box[:, 1]), numpy.min(layover_box[:, 1]))
        max_cols = max(numpy.max(pixel_box[:, 1]), numpy.max(layover_box[:, 1]))

        # determine the padding amount
        padding_fraction = 0.0 if padding_fraction is None else float(padding_fraction)
        if padding_fraction < 0.0:
            padding_fraction = 0.0
        row_pad = max(minimum_pad, padding_fraction*(max_rows-min_rows))
        col_pad = max(minimum_pad, padding_fraction*(max_cols-min_cols))

        # check our bounding information
        rows = sicd.ImageData.NumRows
        cols = sicd.ImageData.NumCols

        chip_rows = [min_rows - row_pad, max_rows + row_pad]
        chip_cols = [min_cols - col_pad, max_cols + col_pad]
        placement = self._check_placement(rows, cols, chip_rows, chip_cols)
        if placement == 3 or (placement == 2 and not populate_in_periphery):
            return placement

        # set the physical data ideal chip size
        physical = PhysicalType.from_ranges(chip_rows, chip_cols, rows, cols)

        # determine nominal shadow vector
        shadow_magnitude = sicd.SCPCOA.ShadowMagnitude
        if shadow_magnitude is None:
            shadow_magnitude = 1.0
        shadow_size = self.Size.Height*shadow_magnitude*magnitude_factor
        shadow_angle = sicd.SCPCOA.Shadow
        shadow_angle = numpy.pi if shadow_angle is None else numpy.deg2rad(shadow_angle)
        shadow_vector = -shadow_size*numpy.array(
            [numpy.cos(shadow_angle)/sicd.Grid.Row.SS, -numpy.sin(shadow_angle)/sicd.Grid.Col.SS])
        shadow_box = pixel_box + shadow_vector

        min_rows = min(min_rows, numpy.min(shadow_box[:, 0]))
        max_rows = max(max_rows, numpy.max(shadow_box[:, 0]))
        min_cols = min(min_cols, numpy.min(shadow_box[:, 1]))
        max_cols = max(max_cols, numpy.max(shadow_box[:, 1]))

        chip_rows = [min_rows - row_pad, max_rows + row_pad]
        chip_cols = [min_cols - col_pad, max_cols + col_pad]
        # set the physical with shadows data ideal chip size
        physical_with_shadows = PhysicalType.from_ranges(chip_rows, chip_cols, rows, cols)

        self.SlantPlane = PlanePhysicalType(
            Physical=physical,
            PhysicalWithShadows=physical_with_shadows)
        return placement

    def get_image_geometry_object_for_sicd(self, include_chip=False):
        """
        Gets the geometry element describing the image geometry for a sicd.

        Returns
        -------
        geometry : Geometry
            The geometry object
        geometry_properties : List[GeometryProperties]
            The associated geometry properties list
        """

        if self.ImageLocation is None:
            raise ValueError('No ImageLocation defined.')

        image_geometry_object, geometry_properties = self.ImageLocation.get_geometry_object()
        if image_geometry_object is None:
            return None, None
        if not include_chip or self.SlantPlane is None:
            return image_geometry_object, geometry_properties

        center_pixel = self.SlantPlane.Physical.CenterPixel.get_array()
        chip_size = self.SlantPlane.Physical.ChipSize.get_array()
        shift = numpy.array([[-0.5, -0.5], [-0.5, 0.5], [0.5, 0.5], [0.5, -0.5]], dtype='float64')
        shift[:, 0] *= chip_size[0]
        shift[:, 1] *= chip_size[1]
        chip_rect = center_pixel + shift
        chip_area = Polygon(coordinates=[chip_rect, ])
        geometry_properties.append(GeometryProperties(name='Physical', color='red'))
        if isinstance(image_geometry_object, GeometryCollection):
            image_geometry_object.geometries.append(chip_area)
        else:
            image_geometry_object = GeometryCollection(geometries=[image_geometry_object, chip_area])

        center_pixel = self.SlantPlane.PhysicalWithShadows.CenterPixel.get_array()
        chip_size = self.SlantPlane.PhysicalWithShadows.ChipSize.get_array()
        shift = numpy.array([[-0.5, -0.5], [-0.5, 0.5], [0.5, 0.5], [0.5, -0.5]], dtype='float64')
        shift[:, 0] *= chip_size[0]
        shift[:, 1] *= chip_size[1]
        chip_rect = center_pixel + shift
        chip_area = Polygon(coordinates=[chip_rect, ])
        geometry_properties.append(GeometryProperties(name='PhysicalWithShadows', color='magenta'))
        image_geometry_object.geometries.append(chip_area)
        return image_geometry_object, geometry_properties

    def add_articulation(self, value):
        if value is None:
            return

        if isinstance(value, str):
            value = StringWithComponentType(Value=value)
        elif isinstance(value, dict):
            value = StringWithComponentType(**value)

        if not isinstance(value, StringWithComponentType):
            raise TypeError('values for Articulation must be of type str or StringWithComponentType')
        
        if self.Articulation is None:
            self.Articulation = [value, ]
        else:
            self.Articulation.append(value)

    def add_configuration(self, value):
        if value is None:
            return

        if isinstance(value, str):
            value = StringWithComponentType(Value=value)
        elif isinstance(value, dict):
            value = StringWithComponentType(**value)

        if not isinstance(value, StringWithComponentType):
            raise TypeError('values for Configuration must be of type str or StringWithComponentType')
        
        if self.Configuration is None:
            self.Configuration = [value, ]
        else:
            self.Configuration.append(value)


# the main type

class ObjectInfoType(Serializable):
    _fields = (
        'NumberOfObjectsInImage', 'NumberOfObjectsInScene', 'LabelSource', 'Objects')
    _required = ('NumberOfObjectsInImage', 'NumberOfObjectsInScene', 'LabelSource', 'Objects')
    _collections_tags = {'Objects': {'array': False, 'child_tag': 'Object'}}
    # descriptors
    NumberOfObjectsInImage = IntegerDescriptor(
        'NumberOfObjectsInImage', _required, strict=DEFAULT_STRICT,
        docstring='Number of ground truthed objects in the image.')  # type: int
    NumberOfObjectsInScene = IntegerDescriptor(
        'NumberOfObjectsInScene', _required, strict=DEFAULT_STRICT,
        docstring='Number of ground truthed objects in the scene.')  # type: int
    LabelSource = SerializableDescriptor(
        'LabelSource', LabelSourceType, _required, strict=DEFAULT_STRICT,
        docstring='The source of the labels')  # type: LabelSourceType
    Objects = SerializableListDescriptor(
        'Objects', TheObjectType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The object collection')  # type: List[TheObjectType]

    def __init__(self, NumberOfObjectsInImage=None, NumberOfObjectsInScene=None,
                 LabelSource=None, Objects=None, **kwargs):
        """
        Parameters
        ----------
        NumberOfObjectsInImage : int
        NumberOfObjectsInScene : int
        LabelSource : LabelSourceType
        Objects : List[ObjectType]
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.NumberOfObjectsInImage = NumberOfObjectsInImage
        self.NumberOfObjectsInScene = NumberOfObjectsInScene
        self.LabelSource = LabelSource
        self.Objects = Objects
        super(ObjectInfoType, self).__init__(**kwargs)

    def set_image_location_from_sicd(
            self, sicd, layover_shift=False, populate_in_periphery=False,
            include_out_of_range=False, padding_fraction=None, minimum_pad=None):
        """
        Set the image location information with respect to the given SICD,
        assuming that the physical coordinates are populated. The `NumberOfObjectsInImage`
        will be set, and `NumberOfObjectsInScene` will be left unchanged.

        Parameters
        ----------
        sicd : SICDType
        layover_shift : bool
            Account for possible layover shift in calculated chip sizes?
        populate_in_periphery : bool
            Populate image information for objects on the periphery?
        include_out_of_range : bool
            Include the objects which are out of range (with no image location information)?
        padding_fraction : None|float
        minimum_pad : None|int|float
        """

        def update_object(temp_object, in_image_count):
            status = temp_object.set_image_location_from_sicd(
                sicd, populate_in_periphery=populate_in_periphery)
            use_object = False
            if status == 0:
                raise ValueError('Object already has image details set')
            if status == 1 or (status == 2 and populate_in_periphery):
                use_object = True
                temp_object.set_chip_details_from_sicd(
                    sicd, layover_shift=layover_shift, populate_in_periphery=True,
                    padding_fraction=padding_fraction, minimum_pad=minimum_pad)
                in_image_count += 1
            return use_object, in_image_count

        objects_in_image = 0
        if include_out_of_range:
            # the objects list is just modified in place
            for the_object in self.Objects:
                _, objects_in_image = update_object(the_object, objects_in_image)
        else:
            # we make a new objects list
            objects = []
            for the_object in self.Objects:
                use_this_object, objects_in_image = update_object(the_object, objects_in_image)
                if use_this_object:
                    objects.append(the_object)
            self.Objects = objects
        self.NumberOfObjectsInImage = objects_in_image



"""
Stateful functions for use in format operations for data segments.

This module introduced in version 1.3.0.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Union, Tuple, Optional

import numpy

from sarpy.io.general.slice_parsing import get_subscript_result_size

logger = logging.getLogger(__name__)


#######
# slice helper functions

def reformat_slice(
        sl_in: slice,
        limit_in: int,
        mirror: bool) -> slice:
    """
    Reformat the slice, with optional reverse operation.

    Note that the mirror operation doesn't run the slice backwards across the
    same elements, but rather creates a mirror image of the slice. This is
    to properly accommodate the data segment reverse symmetry transform.

    Parameters
    ----------
    sl_in : slice
        From prior processing, it is expected that `sl_in.step` is populated,
        and `sl_in.start` is non-negative, and `sl_in.stop` is non-negative or
        `None` (only in th event that `sl_in.step < 0`.
    limit_in : int
        The upper limit for the axis to which this slice pertains.
    mirror : bool
        Create the mirror image slice?

    Returns
    -------
    slice
    """

    if sl_in.step is None:
        raise ValueError('input slice has unpopulated step value')
    if sl_in.start is not None and sl_in.start < 0:
        raise ValueError('input slice has negative start value')
    if sl_in.stop is not None and sl_in.stop < 0:
        raise ValueError('input slice has negative stop value')

    if mirror:
        # make the mirror image of the slice, the step maintains the same sign,
        # and will be reversed by the format function
        if sl_in.step > 0:
            start_in = 0 if sl_in.start is None else sl_in.start
            stop_in = limit_in if sl_in.stop is None else sl_in.stop
            if sl_in.step > (stop_in - start_in):
                step_in = stop_in - start_in
            else:
                step_in = sl_in.step

            # what is the last included location?
            count = int((stop_in - start_in)/float(step_in))
            final_location = start_in + count*step_in
            return slice(limit_in - final_location, limit_in - start_in, step_in)
        else:
            start_in = limit_in - 1 if sl_in.start is None else sl_in.start
            stop_in = -1 if sl_in.stop is None else sl_in.stop

            if sl_in.step < (stop_in - start_in):
                step_in = stop_in - start_in
            else:
                step_in = sl_in.step
            count = int((stop_in - start_in) / float(step_in))
            final_location = start_in + count*step_in
            return slice(limit_in - final_location, limit_in - start_in, step_in)
    else:
        return sl_in


#########
# format function implementations

class FormatFunction(object):
    """
    Stateful function for data orientation and formatting operations associated
    with reading data. *This is specifically intended for use in conjunction
    with `DataSegment`.*

    This allows mapping from raw data to formatted data, for reading data from
    a file and converting it to the form of intended use.

    If the reverse process is implemented, it enables converting from formatted
    data to raw data, for taking common use data and converting it to the raw
    form, for writing data to a file.

    Introduced in version 1.3.0.
    """

    has_inverse = False
    """
    Indicates whether this format function has the inverse call implemented.
    """

    __slots__ = ('_raw_shape', '_formatted_shape', '_reverse_axes', '_transpose_axes', '_reverse_transpose_axes')

    def __init__(
            self,
            raw_shape: Optional[Tuple[int, ...]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Tuple[int, ...]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None):
        """

        Parameters
        ----------
        raw_shape : None|Tuple[int, ...]
        formatted_shape : None|Tuple[int, ...]
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]
        """

        self._raw_shape = None
        self._formatted_shape = None
        self._reverse_axes = None
        self._transpose_axes = None
        self._reverse_transpose_axes = None
        self.set_raw_shape(raw_shape)
        self.set_formatted_shape(formatted_shape)
        self.set_reverse_axes(reverse_axes)
        self.set_transpose_axes(transpose_axes)

    @property
    def raw_shape(self) -> Optional[Tuple[int, ...]]:
        """
        None|Tuple[int, ...]: The expected full possible raw shape.
        """

        return self._raw_shape

    def set_raw_shape(self, value: Optional[Tuple[int, ...]]) -> None:
        if self._raw_shape is not None:
            if value is None or value != self._raw_shape:
                raise ValueError('raw_shape is read only once set')
            return  # nothing to be done
        self._raw_shape = value

    @property
    def raw_ndim(self) -> int:
        if self.raw_shape is None:
            raise ValueError('raw_shape must be set')
        return len(self._raw_shape)

    @property
    def formatted_shape(self) -> Optional[Tuple[int, ...]]:
        """
        None|Tuple[int, ...]: The expected output shape basis.
        """

        return self._formatted_shape

    def set_formatted_shape(self, value: Optional[Tuple[int, ...]]) -> None:
        if self._formatted_shape is not None:
            if value is None or value != self._formatted_shape:
                raise ValueError('formatted_shape is read only once set')
            return  # nothing to be done
        self._formatted_shape = value

    @property
    def formatted_ndim(self) -> int:
        if self.formatted_shape is None:
            raise ValueError('formatted_shape must be set')
        return len(self._formatted_shape)

    @property
    def reverse_axes(self) -> Optional[Tuple[int, ...]]:
        """
        None|Tuple[int, ...]: The collection of axes (with respect to raw order)
        along which we will reverse as part of transformation to output data order.
        If not `None`, then this will be a tuple in strictly increasing order.
        """

        return self._reverse_axes

    def set_reverse_axes(self, value: Optional[Tuple[int, ...]]) -> None:
        if self._reverse_axes is not None:
            if value is None or value != self._reverse_axes:
                raise ValueError('reverse_axes is read only once set')
            return  # nothing to be done
        self._reverse_axes = value

    @property
    def transpose_axes(self) -> Tuple[int, ...]:
        """
        None|Tuple[int, ...]: The transpose order for switching from raw order to
        output order, prior to applying any format function.
        """

        return self._transpose_axes

    def set_transpose_axes(self, value: Optional[Tuple[int, ...]]) -> None:
        if self._transpose_axes is not None:
            if value is None or value != self._transpose_axes:
                raise ValueError('transpose_axes is read only once set')
            return  # nothing to be done
        if value is None:
            return  # nothing to be done

        self._transpose_axes = value
        # inverts the transpose axes mapping
        self._reverse_transpose_axes = tuple([value.index(i) for i in range(len(value))])

    def _get_populated_transpose_axes(self) -> Tuple[int, ...]:
        trans_axes = tuple(range(len(self.raw_shape))) if self.transpose_axes is None else \
            self.transpose_axes
        return trans_axes

    def _verify_shapes_set(self) -> None:
        if self.raw_shape is None or self.formatted_shape is None:
            raise ValueError('raw_shape and formatted_shape must both be set.')

    def _reverse_and_transpose(
            self,
            array: numpy.ndarray,
            inverse=False) -> numpy.ndarray:
        """
        Performs the reverse and transpose operations. This applies to data in raw
        format.

        Parameters
        ----------
        array : numpy.ndarray
        inverse : bool
            If `True`, then this should be the opposite operation.

        Returns
        -------
        numpy.ndarray
        """

        if array.ndim != self.raw_ndim:
            raise ValueError('Got unexpected raw data shape')

        if inverse:
            if self.transpose_axes is not None:
                # NB: this requires a copy, if not trivial
                array = numpy.transpose(array, axes=self._reverse_transpose_axes)
            if self.reverse_axes is not None:
                # NB: these are simply view operations
                for index in self.reverse_axes:
                    array = numpy.flip(array, axis=index)
        else:
            if self.reverse_axes is not None:
                # NB: these are simply view operations
                for index in self.reverse_axes:
                    array = numpy.flip(array, axis=index)
            if self.transpose_axes is not None:
                # NB: this requires a copy, if not trivial
                array = numpy.transpose(array, axes=self.transpose_axes)
        return array

    def __call__(
            self,
            array: numpy.ndarray,
            subscript: Tuple[slice, ...],
            squeeze=True) -> numpy.ndarray:
        """
        Performs the reformatting operation. The output data will have
        dimensions of size 1 squeezed by this operation, it should not generally
        be done before.

        Parameters
        ----------
        array : numpy.ndarray
            The input raw array.
        subscript : Tuple[slice, ...]
            The slice definition which yielded the input raw array.
        squeeze : bool
            Apply numpy.squeeze operation, which eliminates dimensions of size 1?

        Returns
        -------
        numpy.ndarray
            The output formatted array.
        """

        array = self._reverse_and_transpose(array, inverse=False)
        array = self._forward_functional_step(array, subscript)
        if squeeze:
            return numpy.squeeze(array)
        else:
            return array

    def inverse(
            self,
            array: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        """
        Inverse operation which takes in formatted data, and returns
        corresponding raw data.

        Parameters
        ----------
        array : numpy.ndarray
            The input formatted data.
        subscript : Tuple[slice, ...]
            The slice definition which yielded the formatted data.

        Returns
        -------
        numpy.ndarray

        Raises
        ------
        ValueError
            A value error should be raised if `inverse=True` and
            `has_inverse=False`.
        """

        if not self.has_inverse:
            raise ValueError('has_inverse is False')

        array = self._reverse_functional_step(array, subscript)
        array = self._reverse_and_transpose(array, inverse=True)
        return array

    def validate_shapes(self) -> None:
        """
        Validates that the provided `raw_shape` and `formatted_shape` are sensible.

        This should be called only after setting the appropriate values for the
        `raw_shape`, `formatted_shape`, `reverse_axes` and `transpose_axes` properties.

        Raises
        ------
        ValueError
            Raises a ValueError if the shapes are not compatible according to this
            function and the transpose axes argument.
        """

        raise NotImplementedError

    def transform_formatted_slice(
            self,
            subscript: Tuple[slice, ...]) -> Tuple[slice, ...]:
        """
        Transform from the subscript definition in formatted coordinates to
        subscript definition with respect to raw coordinates.

        Parameters
        ----------
        subscript : Tuple[slice, ...]

        Returns
        -------
        Tuple[slice, ...]

        Raises
        ------
        ValueError
            Raised if the desired requirement cannot be met.
        """

        raise NotImplementedError

    def transform_raw_slice(
            self,
            subscript: Tuple[slice, ...]) -> Tuple[slice, ...]:
        """
        Transform from the subscript definition in raw coordinates to
        subscript definition with respect to formatted coordinates.

        Parameters
        ----------
        subscript : Tuple[slice, ...]

        Returns
        -------
        Tuple[slice, ...]

        Raises
        ------
        ValueError
            Raised if the desired requirement cannot be met.
        """

        raise NotImplementedError

    def _forward_functional_step(
            self,
            array: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        """
        Performs the functional operation. This should perform on raw data following
        the reorientation operations provided by :func:`_reverse_and_transpose`.

        Parameters
        ----------
        array : numpy.ndarray
            The raw data to be transformed.
        subscript : Tuple[int, ...]
            The subscript in raw coordinates which would yield the raw data.

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError

    # noinspection PyTypeChecker
    def _reverse_functional_step(
            self,
            array: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        """
        Performs the reverse functional operation. This should perform on formatted data,
        followed by the reorientation operations provided by :func:`_reverse_and_transpose`.

        Parameters
        ----------
        array : numpy.ndarray
            The formatted data to be inverted.
        subscript : Tuple[slice, ...]
            The subscript in formatted coordinates which would yield the formatted data.

        Returns
        -------
        numpy.ndarray
        """

        if not self.has_inverse:
            raise ValueError('has_inverse is False')

        raise NotImplementedError


class IdentityFunction(FormatFunction):
    """
    A format function allowing only reversing and transposing operations, the
    actual functional step is simply the identity function.

    Introduced in version 1.3.0.
    """
    has_inverse = True

    def validate_shapes(self) -> None:
        self._verify_shapes_set()
        if self.raw_ndim != self.formatted_ndim:
            raise ValueError('raw_shape and formatted_shape must have the same length ')

        trans_axes = self._get_populated_transpose_axes()
        if self.raw_ndim != len(trans_axes):
            raise ValueError('raw_shape and transpose_axes must have the same length ')

        # we should have formatted_shape[i] == raw_shape[trans_axes[i]]
        expected_formatted_shape = tuple([self.raw_shape[index] for index in trans_axes])
        if expected_formatted_shape != self.formatted_shape:
            raise ValueError(
                'Input_shape `{}` and transpose_axes `{}` yields expected output shape `{}`\n\t'
                'got formatted_shape `{}`'.format(
                    self.raw_shape, self.transpose_axes, expected_formatted_shape, self.formatted_shape))

    def transform_formatted_slice(
            self,
            subscript: Tuple[slice, ...]) -> Tuple[slice, ...]:
        if len(subscript) != self.formatted_ndim:
            raise ValueError('The length of subscript and formatted_shape must match')

        reverse_axes = () if self.reverse_axes is None else self.reverse_axes
        rev_transpose_axes = tuple(range(len(self.raw_shape))) if self.transpose_axes is None else \
            self._reverse_transpose_axes

        # we will reorder from formatted order into raw order, using the opposite
        # of the transpose axes definition, reversing any axes required according
        # to reverse_axes definition (in raw order)
        out = []
        for i, index in enumerate(rev_transpose_axes):
            # formatted order @ index corresponds to raw order @ i
            rev = (i in reverse_axes)
            shape_limit = self.raw_shape[i]  # also self.formatted_shape[index]
            out.append(reformat_slice(subscript[index], shape_limit, rev))
        return tuple(out)

    def transform_raw_slice(
            self,
            subscript: Tuple[slice, ...]) -> Tuple[slice, ...]:
        if len(subscript) != self.raw_ndim:
            raise ValueError('The length of subscript and raw_shape must match')

        reverse_axes = () if self.reverse_axes is None else self.reverse_axes
        transpose_axes = tuple(range(len(self.formatted_shape))) if self.transpose_axes is None else \
            self.transpose_axes

        # we will reorder from raw order into formatted order, using the transpose
        # axes definition, reversing any axes required according to reverse_axes
        # definition (in raw order)
        out = []
        for i, index in enumerate(transpose_axes):
            # raw order @ index corresponds to formatted order @ i
            rev = (index in reverse_axes)
            shape_limit = self.formatted_shape[i]  # also self.raw_shape[index]
            out.append(reformat_slice(subscript[index], shape_limit, rev))
        return tuple(out)

    def _forward_functional_step(
            self,
            array: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        # the only operations are reordering/reversing, performed by _reverse_and_transpose
        return array

    def _reverse_functional_step(
            self,
            array: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        return array


class ComplexFormatFunction(FormatFunction):
    """
    Reformats data from real/imaginary dimension pairs to complex64 output,
    assuming that the raw data has fixed dimensionality and the real/imaginary
    pairs fall along a given band dimension.

    Introduced in version 1.3.0.
    """
    has_inverse = True
    _allowed_ordering = ('IQ', 'QI', 'MP', 'PM')

    __slots__ = (
        '_band_dimension', '_order', '_raw_dtype')

    def __init__(
            self,
            raw_dtype: Union[str, numpy.dtype],
            order: str,
            raw_shape: Optional[Tuple[int, ...]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Tuple[int, ...]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            band_dimension: int = -1):
        """

        Parameters
        ----------
        raw_dtype : str|numpy.dtype
            The raw datatype. Valid options dependent on the value of order.
        order : str
            One of `('IQ', 'QI', 'MP', 'PM')`. The options `('IQ', 'QI')` allow
            raw_dtype `('int8', 'int16', 'int32', 'float16', 'float32', 'float64')`. The
            options `('MP', 'PM')` allow raw_dtype
            `('uint8', 'uint16', 'uint32', 'float16', 'float32', 'float64')`.
        raw_shape : None|Tuple[int, ...]
        formatted_shape : None|Tuple[int, ...]
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]
        band_dimension : int
            Which band is the complex dimension, **after** the transpose operation.
        """

        self._raw_dtype = numpy.dtype(raw_dtype)  # type: numpy.dtype
        self._band_dimension = None
        self._order = None
        self._set_order(order)

        FormatFunction.__init__(
            self, raw_shape=raw_shape, formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes)
        self._set_band_dimension(band_dimension)

    def set_raw_shape(self, value: Optional[Tuple[int, ...]]) -> None:
        FormatFunction.set_raw_shape(self, value)
        if self._band_dimension is not None:
            self._set_band_dimension(self._band_dimension)

    @property
    def band_dimension(self) -> int:
        """
        int: The band dimension, in raw data after the transpose operation.
        """

        return self._band_dimension

    def _set_band_dimension(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError('band_dimension must be an integer')

        if self._raw_shape is None:
            self._band_dimension = value
            return

        if not (-self.raw_ndim <= value < self.raw_ndim):
            raise ValueError('band_dimension out of bounds.')

        if value < 0:
            value = value + self.raw_ndim

        if self._band_dimension is not None:
            if ((value - self._band_dimension) % self.raw_ndim) != 0:
                raise ValueError('band_dimension is read only once set')
        self._band_dimension = value

    @property
    def order(self) -> str:
        """
        str: The order string, once of `('IQ', 'QI', 'MP', 'PM')`.
        """

        return self._order

    def _set_order(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError('order must be an string')

        value = value.strip().upper()
        if value not in self._allowed_ordering:
            raise ValueError(
                'Order is required to be one of {},\n\t'
                'got `{}`'.format(self._allowed_ordering, value))
        if self._order is not None:
            if value != self._order:
                raise ValueError('order is read only once set')
        self._order = value
        if self._order in ['IQ', 'QI']:
            if self._raw_dtype.name not in [
                    'int8', 'int16', 'int32', 'float16', 'float32', 'float64']:
                raise ValueError(
                    'order is {}, and raw_dtype ({}, {}) must be one of '
                    'int8, int16, int32, float16, float32, or float64'.format(
                        self._order, self._raw_dtype, self._raw_dtype.name))
        elif self._order in ['MP', 'PM']:
            if self._raw_dtype.name not in [
                    'uint8', 'uint16', 'uint32', 'float16', 'float32', 'float64']:
                raise ValueError(
                    'order is {}, and raw_dtype must be one of '
                    'uint8, uint16, uint32, float16, float32, or float64'.format(
                        self._order))
        else:
            raise ValueError('Got unhandled ordering value `{}`'.format(
                self._order))

    def validate_shapes(self) -> None:
        self._verify_shapes_set()
        self._set_band_dimension(self._band_dimension)
        trans_axes = self._get_populated_transpose_axes()
        if self.raw_ndim != len(trans_axes):
            raise ValueError('raw_shape and transpose_axes must have the same length ')

        arranged_shape = tuple([self.raw_shape[index] for index in trans_axes])
        if (arranged_shape[self.band_dimension] % 2) != 0:
            raise ValueError(
                'Input_shape `{}`, transpose_axes `{}` yields rearranged shape `{}`\n\t'
                'entry in band_dimension `{}` should be even'.format(
                    self.raw_shape, self.transpose_axes, arranged_shape, self.band_dimension))
        after_mapping_shape = [entry for entry in arranged_shape]
        after_mapping_shape[self.band_dimension] = int(after_mapping_shape[self.band_dimension]/2)
        after_mapping_shape = tuple(after_mapping_shape)

        if self.raw_ndim == self.formatted_ndim:
            if after_mapping_shape != self.formatted_shape:
                raise ValueError(
                    'Input_shape `{}`, transpose_axes `{}`, band dimension `{}` '
                    'yields expected output shape `{}`\n\t'
                    'got formatted_shape `{}`'.format(
                        self.raw_shape, self.transpose_axes, self.band_dimension,
                        after_mapping_shape, self.formatted_shape))
        elif self.raw_ndim == self.formatted_ndim + 1:
            reduced_shape = [entry for entry in after_mapping_shape]
            reduced_shape.pop(self.band_dimension)
            reduced_shape = tuple(reduced_shape)
            if reduced_shape != self.formatted_shape:
                raise ValueError(
                    'Input_shape `{}`, transpose_axes `{}`, band dimension `{}` '
                    'yields expected output shape `{}`\n\t'
                    'got formatted_shape `{}`'.format(
                        self.raw_shape, self.transpose_axes, self.band_dimension,
                        reduced_shape, self.formatted_shape))
        else:
            raise ValueError(
                'Input_shape `{}`, transpose_axes `{}`, band dimension `{}` '
                'yields expected output shape `{}`\n\t'
                'got formatted_shape `{}`'.format(
                    self.raw_shape, self.transpose_axes, self.band_dimension,
                    arranged_shape, self.formatted_shape))

    def transform_formatted_slice(
            self,
            subscript: Tuple[slice, ...]) -> Tuple[slice, ...]:
        if len(subscript) != len(self.formatted_shape):
            raise ValueError('The length of subscript and formatted_shape must match')

        reverse_axes = () if self.reverse_axes is None else self.reverse_axes
        rev_transpose_axes = tuple(range(len(self.raw_shape))) if self.transpose_axes is None else \
            self._reverse_transpose_axes

        if self.raw_ndim == self.formatted_ndim:
            # there has been no collapse in dimension
            use_subscript = subscript
        else:
            # pad at band dimension (in the order after transpose operation)
            use_subscript = [entry for entry in subscript]
            use_subscript.insert(self.band_dimension, slice(0, 2, 1))

        # we will reorder from formatted order into raw order, using the opposite
        # of the transpose axes definition, reversing any axes required according
        # to reverse_axes definition (in raw order)
        out = []
        for i, index in enumerate(rev_transpose_axes):
            # formatted order @ index corresponds to raw order @ i (possibly padded for missing band dimension)
            rev = (i in reverse_axes)
            shape_limit = self.raw_shape[i]
            if self.raw_ndim == self.formatted_ndim:
                # the band dimension is not flattened, we have to transform
                temp_sl = reformat_slice(use_subscript[index], shape_limit, rev)

                if index == self.band_dimension and temp_sl.step not in [-1, 1]:
                    raise ValueError(
                        'Slicing along the complex dimension and applying this format function\n\t'
                        'is only only permitted using step +/-1')
                if temp_sl.step > 0:
                    start = 2*temp_sl.start if index == self.band_dimension else temp_sl.start
                    # noinspection PyTypeChecker
                    stop = 2*temp_sl.stop if index == self.band_dimension else temp_sl.stop
                    out.append(slice(start, stop, 1))
                elif temp_sl.step < 0:
                    start = 2*temp_sl.start if index == self.band_dimension else temp_sl.start
                    if temp_sl.stop is None:
                        stop = None
                    elif index == self.band_dimension:
                        stop = 2*temp_sl.stop
                    else:
                        stop = temp_sl.stop
                    out.append(slice(start, stop, -1))
            else:
                out.append(reformat_slice(use_subscript[index], shape_limit, rev))
        return tuple(out)

    def transform_raw_slice(
            self,
            subscript: Tuple[slice, ...]) -> Tuple[slice, ...]:
        if len(subscript) != self.raw_ndim:
            raise ValueError('The length of subscript and raw_shape must match')

        reverse_axes = () if self.reverse_axes is None else self.reverse_axes
        transpose_axes = tuple(range(len(self.formatted_shape))) if self.transpose_axes is None else \
            self.transpose_axes

        # we will reorder from raw order into formatted order, using the transpose
        # axes definition, reversing any axes required according to reverse_axes
        # definition (in raw order)
        out = []
        for i, index in enumerate(transpose_axes):
            # raw order @ index corresponds to formatted order @ i
            rev = (index in reverse_axes)
            shape_limit = self.raw_shape[index]  # also self.formatted_shape[i]
            if index == self.band_dimension and self.formatted_ndim < self.raw_ndim:
                # the band dimension has collapsed, so omit anything here
                continue
            else:
                out.append(reformat_slice(subscript[index], shape_limit, rev))
        return tuple(out)

    def _forward_magnitude_theta(
            self,
            data: numpy.ndarray,
            out: numpy.ndarray,
            magnitude: numpy.ndarray,
            theta: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> None:
        if data.dtype.name in ['uint8', 'uint16', 'uint32']:
            bit_depth = data.dtype.itemsize * 8
            theta = theta*2.0*numpy.pi/(1 << bit_depth)
        out.real = magnitude*numpy.cos(theta)
        out.imag = magnitude*numpy.sin(theta)

    def _forward_functional_step(
            self,
            data: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        if data.ndim != self.raw_ndim:
            raise ValueError('Expected raw data of dimension {}'.format(self.raw_ndim))
        if (data.shape[self.band_dimension] % 2) != 0:
            raise ValueError(
                'Requires {} dimensional raw data with even size along dimension {}'.format(
                    self.raw_ndim, self.band_dimension))

        band_dim_size = data.shape[self.band_dimension]
        if self.formatted_ndim < self.raw_ndim:
            out_shape = data.shape[:self.band_dimension] + data.shape[self.band_dimension + 1:]
        else:
            out_shape = data.shape[:self.band_dimension] + \
                        (int(band_dim_size/2), ) + \
                        data.shape[self.band_dimension + 1:]

        out = numpy.empty(out_shape, dtype='complex64')
        if self.order == 'IQ':
            out.real = numpy.reshape(
                data.take(indices=range(0, band_dim_size, 2), axis=self.band_dimension), out.shape)
            out.imag = numpy.reshape(
                data.take(indices=range(1, band_dim_size, 2), axis=self.band_dimension), out.shape)
        elif self.order == 'QI':
            out.imag = numpy.reshape(
                data.take(indices=range(0, band_dim_size, 2), axis=self.band_dimension), out.shape)
            out.real = numpy.reshape(
                data.take(indices=range(1, band_dim_size, 2), axis=self.band_dimension), out.shape)
        elif self.order in ['MP', 'PM']:
            if self.order == 'MP':
                mag = numpy.reshape(
                    data.take(indices=range(0, band_dim_size, 2), axis=self.band_dimension), out.shape)
                theta = numpy.reshape(
                    data.take(indices=range(1, band_dim_size, 2), axis=self.band_dimension), out.shape)
            else:
                mag = numpy.reshape(
                    data.take(indices=range(1, band_dim_size, 2), axis=self.band_dimension), out.shape)
                theta = numpy.reshape(
                    data.take(indices=range(0, band_dim_size, 2), axis=self.band_dimension), out.shape)
            self._forward_magnitude_theta(data, out, mag, theta, subscript)
        else:
            raise ValueError('Unhandled order value {}'.format(self.order))

        return out

    def _reverse_magnitude_theta(
            self,
            data: numpy.ndarray,
            out: numpy.ndarray,
            magnitude: numpy.ndarray,
            theta: numpy.ndarray,
            slice0: Tuple[slice, ...],
            slice1: Tuple[slice, ...]) -> None:
        if self._raw_dtype.name in ['uint8', 'uint16', 'uint32']:
            bit_depth = self._raw_dtype.itemsize * 8
            theta *= (1 << bit_depth) / (2 * numpy.pi)
            theta = numpy.round(theta)
            magnitude = numpy.round(magnitude)

        if self.order == 'MP':
            out[slice0] = magnitude
            out[slice1] = theta
        else:
            out[slice1] = magnitude
            out[slice0] = theta

    def _reverse_functional_step(
            self,
            data: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        if data.ndim != self.formatted_ndim:
            raise ValueError('Expected formatted data of dimension {}'.format(self.formatted_ndim))

        if self.formatted_ndim < self.raw_ndim:
            out_shape = data.shape[:self.band_dimension] + (2,) + data.shape[self.band_dimension:]
            use_shape = data.shape[:self.band_dimension] + (1,) + data.shape[self.band_dimension:]
        else:
            band_dim_size = data.shape[self.band_dimension]
            out_shape = data.shape[:self.band_dimension] + \
                (2*band_dim_size, ) + \
                data.shape[self.band_dimension + 1:]
            use_shape = data.shape[:self.band_dimension] + \
                (band_dim_size, ) + \
                data.shape[self.band_dimension + 1:]

        slice0 = []
        slice1 = []
        for index, siz in enumerate(out_shape):
            if index == self.band_dimension:
                slice0.append(slice(0, siz, 2))
                slice1.append(slice(1, siz, 2))
            else:
                slice0.append(slice(0, siz, 1))
                slice1.append(slice(0, siz, 1))
        slice0 = tuple(slice0)
        slice1 = tuple(slice1)

        out = numpy.empty(out_shape, dtype=self._raw_dtype)
        if self.order == 'IQ':
            out[slice0] = numpy.reshape(data.real, use_shape)
            out[slice1] = numpy.reshape(data.imag, use_shape)
        elif self.order == 'QI':
            out[slice1] = numpy.reshape(data.real, use_shape)
            out[slice0] = numpy.reshape(data.imag, use_shape)
        elif self.order in ['MP', 'PM']:
            magnitude = numpy.reshape(numpy.abs(data), use_shape)
            theta = numpy.reshape(numpy.arctan2(data.imag, data.real), use_shape)
            theta[theta < 0] += 2*numpy.pi
            self._reverse_magnitude_theta(data, out, magnitude, theta, slice0, slice1)
        else:
            raise ValueError('Unhandled order value {}'.format(self.order))
        return out


class SingleLUTFormatFunction(FormatFunction):
    """
    Reformat the raw data according to the use of a single 8-bit lookup table.
    In the case of a 2-d LUT, and effort to slice on the final dimension
    (from the LUT) is not supported.

    Introduced in version 1.3.0.
    """

    has_inverse = False

    __slots__ = ('_lookup_table', )

    def __init__(
            self,
            lookup_table: numpy.ndarray,
            raw_shape: Optional[Tuple[int, ...]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Tuple[int, ...]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None):
        """

        Parameters
        ----------
        lookup_table : numpy.ndarray
            The 8-bit lookup table.
        raw_shape : None|Tuple[int, ...]
        formatted_shape : None|Tuple[int, ...]
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]
        """

        self._lookup_table = None
        if not isinstance(lookup_table, numpy.ndarray):
            raise ValueError('requires a numpy.ndarray, got {}'.format(type(lookup_table)))
        if lookup_table.dtype.name != 'uint8':
            raise ValueError('requires a numpy.ndarray of uint8 dtype, got {}'.format(lookup_table.dtype))
        if lookup_table.ndim == 2 and lookup_table.shape[1] == 1:
            lookup_table = numpy.reshape(lookup_table, (-1, ))
        self._lookup_table = lookup_table

        FormatFunction.__init__(
            self, raw_shape=raw_shape, formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes)

    @property
    def lookup_table(self) -> numpy.ndarray:
        return self._lookup_table

    def validate_shapes(self) -> None:
        self._verify_shapes_set()
        trans_axes = self._get_populated_transpose_axes()
        if self.raw_ndim != len(trans_axes):
            raise ValueError('raw_shape and transpose_axes must have the same length')

        arranged_shape = [self.raw_shape[index] for index in trans_axes]
        if self.lookup_table.ndim == 2:
            arranged_shape.append(self.lookup_table.shape[1])
        arranged_shape = tuple(arranged_shape)

        if arranged_shape != self.formatted_shape:
            raise ValueError(
                'Input_shape `{}`, transpose_axes `{}` and lookup table\n\t'
                'yields expected output shape `{}` got formatted_shape `{}`'.format(
                    self.raw_shape, self.transpose_axes, arranged_shape, self.formatted_shape))

    def transform_formatted_slice(
            self,
            subscript: Tuple[slice, ...]) -> Tuple[slice, ...]:
        if len(subscript) != self.formatted_ndim:
            raise ValueError('The length of subscript and formatted_shape must match')

        reverse_axes = () if self.reverse_axes is None else self.reverse_axes
        rev_transpose_axes = tuple(range(len(self.raw_shape))) if self.transpose_axes is None else \
            self._reverse_transpose_axes

        # we will reorder from formatted order into raw order, using the opposite
        # of the transpose axes definition, reversing any axes required according
        # to reverse_axes definition (in raw order)
        out = []
        # NB: for 2-d LUT, the final slice will be ignored here (as it should)
        for i, index in enumerate(rev_transpose_axes):
            # formatted order @ index corresponds to raw order @ i
            rev = (i in reverse_axes)
            shape_limit = self.raw_shape[i]  # also self.formatted_shape[index]
            out.append(reformat_slice(subscript[index], shape_limit, rev))
        return tuple(out)

    def transform_raw_slice(
            self,
            subscript: Tuple[slice, ...]) -> Tuple[slice, ...]:
        if len(subscript) != self.raw_ndim:
            raise ValueError('The length of subscript and raw_shape must match')

        reverse_axes = () if self.reverse_axes is None else self.reverse_axes
        transpose_axes = tuple(range(len(self.formatted_shape))) if self.transpose_axes is None else \
            self.transpose_axes

        # we will reorder from raw order into formatted order, using the transpose
        # axes definition, reversing any axes required according to reverse_axes
        # definition (in raw order)
        out = []
        for i, index in enumerate(transpose_axes):
            # raw order @ index corresponds to formatted order @ i
            rev = (index in reverse_axes)
            shape_limit = self.formatted_shape[index]  # also self.raw_shape[i]
            out.append(reformat_slice(subscript[index], shape_limit, rev))
        if self.raw_ndim < self.formatted_ndim:
            # 2-d lookup table
            lim = self.formatted_shape[-1]
            out.append(slice(0, lim, 1))
        return tuple(out)

    def _forward_functional_step(
            self,
            array: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> numpy.ndarray:
        if not isinstance(array, numpy.ndarray):
            raise ValueError('requires a numpy.ndarray, got {}'.format(type(array)))

        if array.dtype.name not in ['uint8', 'uint16']:
            raise ValueError('requires a numpy.ndarray of uint8 or uint16 dtype, '
                             'got {}'.format(array.dtype.name))

        if array.ndim != 2:
            raise ValueError('Requires a two-dimensional numpy.ndarray, got shape {}'.format(array.shape))
        temp = numpy.reshape(array, (-1, ))
        out = self.lookup_table[temp]
        if self.lookup_table.ndim == 2:
            out_shape = array.shape + (self.lookup_table.shape[1], )
        else:
            out_shape = array.shape
        return numpy.reshape(out, out_shape)

    def __call__(
            self,
            array: numpy.ndarray,
            subscript: Tuple[slice, ...],
            squeeze=True) -> numpy.ndarray:
        array = self._reverse_and_transpose(array, inverse=False)
        array = self._forward_functional_step(array, subscript)
        if self.raw_ndim < self.formatted_ndim:
            # apply slice in the band (final dimension)
            array = array.take(
                indices=numpy.arange(self.formatted_shape[-1])[subscript[-1]], axis=-1)
            # ensure shape is as expected - any squeeze handled consistently
            _, out_shape = get_subscript_result_size(subscript, self.formatted_shape)
            array = numpy.reshape(array, out_shape)
        if squeeze:
            return numpy.squeeze(array)
        else:
            return array




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class IMGE(TREElement):
    def __init__(self, value):
        super(IMGE, self).__init__()
        self.add_field('IIDI', 's', 80, value)
        self.add_field('NPARI', 'd', 2, value)


class DERCOV(TREElement):
    def __init__(self, value):
        super(DERCOV, self).__init__()
        self.add_field('DERCOV', 's', 21, value)


class RSMDCAType(TREElement):
    def __init__(self, value):
        super(RSMDCAType, self).__init__()
        self.add_field('IID', 's', 80, value)
        self.add_field('EDITION', 's', 40, value)
        self.add_field('TID', 's', 40, value)
        self.add_field('NPAR', 'd', 2, value)
        self.add_field('NIMGE', 'd', 3, value)
        self.add_field('NPART', 'd', 5, value)
        self.add_loop('IMGEs', self.NIMGE, IMGE, value)
        self.add_field('XUOL', 's', 21, value)
        self.add_field('YUOL', 's', 21, value)
        self.add_field('ZUOL', 's', 21, value)
        self.add_field('XUXL', 's', 21, value)
        self.add_field('XUYL', 's', 21, value)
        self.add_field('XUZL', 's', 21, value)
        self.add_field('YUXL', 's', 21, value)
        self.add_field('YUYL', 's', 21, value)
        self.add_field('YUZL', 's', 21, value)
        self.add_field('ZUXL', 's', 21, value)
        self.add_field('ZUYL', 's', 21, value)
        self.add_field('ZUZL', 's', 21, value)
        self.add_field('IR0', 's', 2, value)
        self.add_field('IRX', 's', 2, value)
        self.add_field('IRY', 's', 2, value)
        self.add_field('IRZ', 's', 2, value)
        self.add_field('IRXX', 's', 2, value)
        self.add_field('IRXY', 's', 2, value)
        self.add_field('IRXZ', 's', 2, value)
        self.add_field('IRYY', 's', 2, value)
        self.add_field('IRYZ', 's', 2, value)
        self.add_field('IRZZ', 's', 2, value)
        self.add_field('IC0', 's', 2, value)
        self.add_field('ICX', 's', 2, value)
        self.add_field('ICY', 's', 2, value)
        self.add_field('ICZ', 's', 2, value)
        self.add_field('ICXX', 's', 2, value)
        self.add_field('ICXY', 's', 2, value)
        self.add_field('ICXZ', 's', 2, value)
        self.add_field('ICYY', 's', 2, value)
        self.add_field('ICYZ', 's', 2, value)
        self.add_field('ICZZ', 's', 2, value)
        self.add_field('GX0', 's', 2, value)
        self.add_field('GY0', 's', 2, value)
        self.add_field('GZ0', 's', 2, value)
        self.add_field('GXR', 's', 2, value)
        self.add_field('GYR', 's', 2, value)
        self.add_field('GZR', 's', 2, value)
        self.add_field('GS', 's', 2, value)
        self.add_field('GXX', 's', 2, value)
        self.add_field('GXY', 's', 2, value)
        self.add_field('GXZ', 's', 2, value)
        self.add_field('GYX', 's', 2, value)
        self.add_field('GYY', 's', 2, value)
        self.add_field('GYZ', 's', 2, value)
        self.add_field('GZX', 's', 2, value)
        self.add_field('GZY', 's', 2, value)
        self.add_field('GZZ', 's', 2, value)
        self.add_loop('DERCOVs', self.dercov_count(), DERCOV, value)

    def dercov_count(self):  # type: () -> int
        return int((self.NPART+1)*self.NPART/2)


class RSMDCA(TREExtension):
    _tag_value = 'RSMDCA'
    _data_type = RSMDCAType




import numpy

__author__ = "Thomas McCullough"
__classification__ = "UNCLASSIFIED"


def argument_validation(lat, lon):
    if not isinstance(lat, numpy.ndarray):
        lat = numpy.array(lat)
    if not isinstance(lon, numpy.ndarray):
        lon = numpy.array(lon)
    if lat.shape != lon.shape:
        raise ValueError(
            'lat and lon must have the same shape, got '
            'lat.shape = {}, lon.shape = {}'.format(lat.shape, lon.shape))
    o_shape = lat.shape
    lat = numpy.reshape(lat, (-1,))
    lon = numpy.reshape(lon, (-1,))

    return o_shape, lat, lon



"""
Module for reading and writing SIDD files
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from functools import reduce
import re
from typing import List, Tuple, Sequence, Union, BinaryIO, Optional

import numpy

from sarpy.io.xml.base import parse_xml_from_string

from sarpy.io.general.utils import is_file_like
from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.nitf import NITFDetails, NITFReader, NITFWriter, \
    interpolate_corner_points_string, ImageSubheaderManager, \
    GraphicsSubheaderManager, TextSubheaderManager, DESSubheaderManager, \
    RESSubheaderManager, NITFWritingDetails, default_image_segmentation
from sarpy.io.general.nitf_elements.nitf_head import NITFHeader
from sarpy.io.general.nitf_elements.des import DataExtensionHeader, XMLDESSubheader
from sarpy.io.general.nitf_elements.security import NITFSecurityTags
from sarpy.io.general.nitf_elements.image import ImageSegmentHeader, \
    ImageSegmentHeader0, ImageBands, ImageBand

from sarpy.io.product.base import SIDDTypeReader
from sarpy.io.product.sidd3_elements.SIDD import SIDDType as SIDDType3
from sarpy.io.product.sidd2_elements.SIDD import SIDDType as SIDDType2
from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd import extract_clas as extract_clas_sicd, \
    create_security_tags_from_sicd


logger = logging.getLogger(__name__)

########
# module variables
_class_priority = {'U': 0, 'R': 1, 'C': 2, 'S': 3, 'T': 4}
SIDD_TYPES = (SIDDType1, SIDDType2, SIDDType3)


#########
# Helper object for initially parses NITF header - specifically looking for SICD elements


class SIDDDetails(NITFDetails):
    """
    SIDD are stored in NITF 2.1 files.
    """

    __slots__ = (
        '_is_sidd', '_sidd_meta', '_sicd_meta')

    def __init__(self, file_object: Union[str, BinaryIO]):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
            file name or file like object for a NITF 2.1 or 2.0 containing a SIDD
        """

        self._img_headers = None
        self._is_sidd = False
        self._sidd_meta = None
        self._sicd_meta = None
        NITFDetails.__init__(self, file_object)

        if self._nitf_header.ImageSegments.subhead_sizes.size == 0:
            raise SarpyIOError('There are no image segments defined.')
        if self._nitf_header.GraphicsSegments.item_sizes.size > 0:
            raise SarpyIOError('A SIDD file does not allow for graphics segments.')
        if self._nitf_header.DataExtensions.subhead_sizes.size == 0:
            raise SarpyIOError(
                'A SIDD file requires at least one data extension, containing the '
                'SIDD xml structure.')

        # define the sidd and sicd metadata
        self._find_sidd()
        if not self.is_sidd:
            raise SarpyIOError('Could not find SIDD xml data extensions.')

    @property
    def is_sidd(self) -> bool:
        """
        bool: whether file name corresponds to a SIDD file, or not.
        """

        return self._is_sidd

    @property
    def sidd_meta(self) -> Union[SIDDType3, SIDDType2, SIDDType1, List[SIDDType3], List[SIDDType2], List[SIDDType1]]:
        """
        None|SIDDType3|SIDDType2|SIDDType1|List[SIDDType3]|List[SIDDType2]|List[SIDDType1]: the sidd meta-data structure(s).
        """

        return self._sidd_meta

    @property
    def sicd_meta(self) -> Optional[List[SICDType]]:
        """
        None|List[SICDType]: the sicd meta-data structure(s).
        """

        return self._sicd_meta

    def _find_sidd(self) -> None:
        self._is_sidd = False
        if self.des_subheader_offsets is None:
            return

        self._sidd_meta = []
        self._sicd_meta = []

        for i in range(self.des_subheader_offsets.size):
            subhead_bytes = self.get_des_subheader_bytes(i)
            if subhead_bytes.startswith(b'DEXML_DATA_CONTENT'):
                des_bytes = self.get_des_bytes(i)
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SIDD' in root_node.tag:
                        self._is_sidd = True
                        self._sidd_meta.append(SIDDType3.from_node(root_node, xml_ns, ns_key='default'))
                    elif 'SICD' in root_node.tag:
                        self._sicd_meta.append(SICDType.from_node(root_node, xml_ns, ns_key='default'))
                except Exception as e:
                    logger.error('Failed checking des xml header at index {} with error {}'.format(i, e))
                    continue
            elif subhead_bytes.startswith(b'DESIDD_XML'):
                # This is an old format SIDD header
                des_bytes = self.get_des_bytes(i)
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SIDD' in root_node.tag:
                        self._is_sidd = True
                        self._sidd_meta.append(SIDDType3.from_node(root_node, xml_ns, ns_key='default'))
                except Exception as e:
                    logger.error(
                        'We found an apparent old-style SIDD DES header at index {},\n\t'
                        'but failed parsing with error {}'.format(i, e))
                    continue
            elif subhead_bytes.startswith(b'DESICD_XML'):
                # This is an old format SICD header
                des_bytes = self.get_des_bytes(i)
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SICD' in root_node.tag:
                        self._sicd_meta.append(SICDType.from_node(root_node, xml_ns, ns_key='default'))
                except Exception as e:
                    logger.error(
                        'We found an apparent old-style SICD DES header at index {},\n\t'
                        'but failed parsing with error {}'.format(i, e))
                    continue

        if not self._is_sidd:
            return

        for sicd in self._sicd_meta:
            sicd.derive()


#######
#  The actual reading implementation

def _check_iid_format(iid1: str) -> bool:
    if not (iid1[:4] == 'SIDD' and iid1[4:].isnumeric()):
        return False
    return True


class SIDDReader(NITFReader, SIDDTypeReader):
    """
    A reader object for a SIDD file (NITF container with SIDD contents)
    """

    def __init__(self, nitf_details):
        """

        Parameters
        ----------
        nitf_details : str|BinaryIO|SIDDDetails
            filename, file-like object, or SIDDDetails object
        """

        if isinstance(nitf_details, str) or is_file_like(nitf_details):
            nitf_details = SIDDDetails(nitf_details)
        if not isinstance(nitf_details, SIDDDetails):
            raise TypeError('The input argument for SIDDReader must be a filename or '
                            'SIDDDetails object.')

        if not nitf_details.is_sidd:
            raise ValueError(
                'The input file passed in appears to be a NITF 2.1 file that does not contain '
                'valid sidd metadata.')

        self._nitf_details = nitf_details
        SIDDTypeReader.__init__(self, None, self.nitf_details.sidd_meta, self.nitf_details.sicd_meta)
        NITFReader.__init__(self, nitf_details, reader_type="SIDD")
        self._check_sizes()

    @property
    def nitf_details(self) -> SIDDDetails:
        """
        SIDDDetails: The SIDD NITF details object.
        """

        return self._nitf_details

    def _check_image_segment_for_compliance(
            self,
            index: int,
            img_header: Union[ImageSegmentHeader, ImageSegmentHeader0]) -> bool:

        out = NITFReader._check_image_segment_for_compliance(self, index, img_header)
        if not out:
            return out

        out = True
        # skip anything but SAR for now (i.e. legend)
        if img_header.ICAT != 'SAR':
            logger.info('Image segment at index {} has ICAT != SAR'.format(index))
            out = False
        if not _check_iid_format(img_header.IID1):
            logger.info('Image segment at index {} has IID1 {}'.format(index, img_header.IID1))
            out = False
        return out

    def find_image_segment_collections(self) -> Tuple[Tuple[int, ...]]:
        segments = [[] for _ in self._sidd_meta]  # the number of segments that we should have

        for i, img_header in enumerate(self.nitf_details.img_headers):
            if i in self.unsupported_segments:
                continue  # skip these...

            iid1 = img_header.IID1  # required to be of the form SIDD######
            if not _check_iid_format(iid1):
                raise ValueError('Got unhandled image segment at index {}'.format(i))
            element = int(iid1[4:7])
            if element > len(self._sidd_meta):
                raise ValueError(
                    'Got image segment iid1 {}, but there are only {} sidd elements'.format(
                        iid1, len(self._sidd_meta)))
            segments[element - 1].append(i)
        # Ensure that all segments are populated with something...
        for i, entry in enumerate(segments):
            if len(entry) < 1:
                raise ValueError('Did not find any image segments for SIDD {}'.format(i))
        # noinspection PyTypeChecker
        return tuple(tuple(entry) for entry in segments)


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name: Union[str, BinaryIO]) -> Optional[SIDDReader]:
    """
    Tests whether a given file_name corresponds to a SIDD file.
    Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    SIDDReader|None
        `SIDDReader` instance if SIDD file, `None` otherwise
    """

    try:
        nitf_details = SIDDDetails(file_name)
        if nitf_details.is_sidd:
            logger.info('File {} is determined to be a SIDD (NITF format) file.'.format(file_name))
            return SIDDReader(nitf_details)
        else:
            return None
    except SarpyIOError:
        # we don't want to catch parsing errors, for now
        return None


#########
# The writer implementation

def validate_sidd_for_writing(
        sidd_meta: Union[SIDDType3, SIDDType2, SIDDType1, List[SIDDType3], List[SIDDType2],
                   List[SIDDType1]]) -> Union[Tuple[SIDDType3, ...], Tuple[SIDDType2, ...], Tuple[SIDDType1, ...]]:
    """
    Helper method which ensures the provided SIDD structure is appropriate.

    Parameters
    ----------
    sidd_meta : SIDDType3|List[SIDDType3]|SIDDType2|List[SIDDType2]|SIDDType1|List[SIDDType1]

    Returns
    -------
    Tuple[SIDDType3, ...]|Tuple[SIDDType2, ...]|Tuple[SIDDType1, ...]
    """

    def inspect_sidd(the_sidd: Union[SIDDType3, SIDDType2, SIDDType1]) -> None:
        # we must have the image size
        if the_sidd.Measurement is None:
            raise ValueError('The sidd_meta has un-populated Measurement, '
                             'and nothing useful can be inferred.')
        if the_sidd.Measurement.PixelFootprint is None:
            raise ValueError('The sidd_meta has un-populated Measurement.PixelFootprint, '
                             'and this is not valid for writing.')

        # we must have the pixel type
        if the_sidd.Display is None:
            raise ValueError('The sidd_meta has un-populated Display, '
                             'and nothing useful can be inferred.')
        if the_sidd.Display.PixelType is None:
            raise ValueError('The sidd_meta has un-populated Display.PixelType, '
                             'and nothing useful can be inferred.')
        # No support for LUT until necessary
        if the_sidd.Display.PixelType in ('MONO8LU', 'RGB8LU'):
            raise ValueError('PixelType requiring lookup table currently unsupported.')

        # we must have collection time
        if the_sidd.ExploitationFeatures is None:
            raise ValueError('The sidd_meta has un-populated ExploitationFeatures.')
        if the_sidd.ExploitationFeatures.Collections is None or len(the_sidd.ExploitationFeatures.Collections) == 0:
            raise ValueError('The sidd_meta has un-populated ExploitationFeatures.Collections.')
        if the_sidd.ExploitationFeatures.Collections[0].Information is None:
            raise ValueError('The sidd_meta has un-populated ExploitationFeatures.Collections.Information.')
        if the_sidd.ExploitationFeatures.Collections[0].Information.CollectionDateTime is None:
            raise ValueError('The sidd_meta has un-populated ExploitationFeatures.Collections.'
                             'Information.CollectionDateTime')

        # try validating versus the appropriate schema
        xml_str = the_sidd.to_xml_string(tag='SIDD')
        from sarpy.consistency.sidd_consistency import evaluate_xml_versus_schema
        if isinstance(the_sidd, SIDDType1):
            urn = 'urn:SIDD:1.0.0'
        elif isinstance(the_sidd, SIDDType2):
            urn = 'urn:SIDD:2.0.0'
        elif isinstance(the_sidd, SIDDType3):
            urn = 'urn:SIDD:3.0.0'
        else:
            raise ValueError('Unhandled type {}'.format(type(the_sidd)))
        result = evaluate_xml_versus_schema(xml_str, urn)
        if result is False:
            logger.warning(
                'The provided SIDD does not properly validate\n\t'
                'against the schema for {}'.format(urn))

    if isinstance(sidd_meta, (SIDDType3, SIDDType2, SIDDType1)):
        inspect_sidd(sidd_meta)
        # noinspection PyRedundantParentheses
        return (sidd_meta, )
    elif isinstance(sidd_meta, (tuple, list)):
        out = []
        for entry in sidd_meta:
            if not isinstance(entry, (SIDDType3, SIDDType2, SIDDType1)):
                raise TypeError('All entries are required to be an instance of SIDDType, '
                                'got type {}'.format(type(entry)))
            inspect_sidd(entry)
            out.append(entry)
        return tuple(out)
    else:
        raise TypeError('sidd_meta is required to be an instance of SIDDType or a list/tuple '
                        'of such instances, got {}'.format(type(sidd_meta)))


def validate_sicd_for_writing(sicd_meta: Union[SICDType, Sequence[SICDType]]) -> Optional[Tuple[SICDType, ...]]:
    """
    Helper method which ensures the provided SICD structure is appropriate.

    Parameters
    ----------
    sicd_meta : SICDType|List[SICDType]

    Returns
    -------
    None|Tuple[SICDType]
    """

    if sicd_meta is None:
        return None
    if isinstance(sicd_meta, SICDType):
        # noinspection PyRedundantParentheses
        return (sicd_meta, )
    elif isinstance(sicd_meta, (tuple, list)):
        out = []
        for entry in sicd_meta:
            if not isinstance(entry, SICDType):
                raise TypeError('All entries are required to be an instance of SICDType, '
                                'got type {}'.format(type(entry)))
            out.append(entry)
        return tuple(out)
    else:
        raise TypeError('sicd_meta is required to be an instance of SICDType or a list/tuple '
                        'of such instances, got {}'.format(type(sicd_meta)))


def extract_clas(the_sidd: Union[SIDDType3, SIDDType2, SIDDType1]) -> str:
    """
    Extract the classification string from a SIDD as appropriate for NITF Security
    tags CLAS attribute.

    Parameters
    ----------
    the_sidd : SIDDType3|SIDDType2|SIDDType1

    Returns
    -------
    str
    """

    class_str = the_sidd.ProductCreation.Classification.classification

    if class_str is None or class_str == '':
        return 'U'
    else:
        return class_str[:1]


def extract_clsy(the_sidd: Union[SIDDType3, SIDDType2, SIDDType1]) -> str:
    """
    Extract the ownerProducer string from a SIDD as appropriate for NITF Security
    tags CLSY attribute.

    Parameters
    ----------
    the_sidd : SIDDType3|SIDDType2|SIDDType1

    Returns
    -------
    str
    """

    owner = the_sidd.ProductCreation.Classification.ownerProducer.upper()
    if owner is None:
        return ''
    elif owner in ('USA', 'CAN', 'AUS', 'NZL'):
        return owner[:2]
    elif owner == 'GBR':
        return 'UK'
    elif owner == 'NATO':
        return 'XN'
    else:
        logger.warning(
            'Got owner {}, and the CLSY will be truncated\n\t'
            'to two characters.'.format(owner))
        return owner[:2]


def create_security_tags_from_sidd(sidd_meta: Union[SIDDType3, SIDDType2, SIDDType1]) -> NITFSecurityTags:
    def get_basic_args():
        out = {}
        sec_tags = sidd_meta.NITF.get('Security', {})
        # noinspection PyProtectedMember
        for fld in NITFSecurityTags._ordering:
            if fld in sec_tags:
                out[fld] = sec_tags[fld]
        return out

    def get_clas():
        if 'CLAS' in args:
            return
        args['CLAS'] = extract_clas(sidd_meta)

    def get_code():
        # TODO: what to do here?
        if 'CODE' in args:
            return

    def get_clsy():
        if 'CLSY' in args:
            return
        args['CLSY'] = extract_clsy(sidd_meta)

    args = get_basic_args()
    if sidd_meta.ProductCreation is not None and \
            sidd_meta.ProductCreation.Classification is not None:
        get_clas()
        get_code()
        get_clsy()
    return NITFSecurityTags(**args)


class SIDDWritingDetails(NITFWritingDetails):
    __slots__ = (
        '_sidd_meta', '_sicd_meta', '_security_tags',
        '_sidd_security_tags', '_sicd_security_tags',
        '_row_limit')

    def __init__(
            self,
            sidd_meta: Union[SIDDType3, SIDDType2, SIDDType1, Sequence[SIDDType3], Sequence[SIDDType2], Sequence[SIDDType1]],
            sicd_meta: Optional[Union[SICDType, Sequence[SICDType]]],
            row_limit: Optional[int] = None,
            additional_des: Optional[Sequence[DESSubheaderManager]] = None,
            graphics_managers: Optional[Tuple[GraphicsSubheaderManager, ...]] = None,
            text_managers: Optional[Tuple[TextSubheaderManager, ...]] = None,
            res_managers: Optional[Tuple[RESSubheaderManager, ...]] = None):
        """

        Parameters
        ----------
        sidd_meta : SIDDType3|List[SIDDType3]|SIDDType2|List[SIDDType2]|SIDDType1|List[SIDDType1]
        sicd_meta : SICDType
        row_limit : None|int
            Desired row limit for the sicd image segments. Non-positive values
            or values > 99999 will be ignored.
        additional_des : None|Sequence[DESSubheaderManager]
        graphics_managers: Optional[Tuple[GraphicsSubheaderManager, ...]]
        text_managers: Optional[Tuple[TextSubheaderManager, ...]]
        res_managers: Optional[Tuple[RESSubheaderManager, ...]]
        """

        self._sidd_meta = None
        self._sidd_security_tags = None
        self._set_sidd_meta(sidd_meta)

        self._sicd_meta = None
        self._sicd_security_tags = None
        self._set_sicd_meta(sicd_meta)

        self._security_tags = None
        self._create_security_tags()

        self._row_limit = None
        self._set_row_limit(row_limit)

        header = self._create_header()
        image_managers, image_segment_collection, image_segment_coordinates = self._create_image_segments()
        des_managers = self._create_des_segments(additional_des)
        NITFWritingDetails.__init__(
            self,
            header,
            image_managers=image_managers,
            image_segment_collections=image_segment_collection,
            image_segment_coordinates=image_segment_coordinates,
            graphics_managers=graphics_managers,
            text_managers=text_managers,
            des_managers=des_managers,
            res_managers=res_managers)

    @property
    def sidd_meta(self) -> Union[Tuple[SIDDType3, ...], Tuple[SIDDType2, ...], Tuple[SIDDType1, ...]]:
        """
        Tuple[SIDDType3, ...]: The sidd metadata.
        """

        return self._sidd_meta

    def _set_sidd_meta(self, value) -> None:
        if self._sidd_meta is not None:
            raise ValueError('sidd_meta is read only')
        if value is None:
            raise ValueError('sidd_meta is required.')

        self._sidd_meta = validate_sidd_for_writing(value)
        # noinspection PyTypeChecker
        self._sidd_security_tags = tuple(create_security_tags_from_sidd(entry) for entry in self._sidd_meta)

    @property
    def sicd_meta(self) -> Tuple[SICDType, ...]:
        """
        Tuple[SICDType, ...]: The sicd metadata
        """

        return self._sicd_meta

    def _set_sicd_meta(self, value) -> None:
        if self._sicd_meta is not None:
            raise ValueError('sicd_meta is read only')
        if value is None:
            self._sicd_meta = None
            self._sicd_security_tags = None
            return

        self._sicd_meta = validate_sicd_for_writing(value)
        # noinspection PyTypeChecker
        self._sicd_security_tags = tuple(create_security_tags_from_sicd(entry) for entry in self._sicd_meta)

    @property
    def row_limit(self) -> Tuple[int, ...]:
        return self._row_limit

    def _set_row_limit(self, value) -> None:
        if value is not None:
            if not isinstance(value, int):
                raise TypeError('row_bounds must be an integer')
            if value < 1:
                value = None

        if value is None or value > 99999:
            value = 99999

        im_seg_limit = 10**10 - 2  # allowable image segment size
        row_limit = []
        for sidd in self.sidd_meta:
            row_memory_size = sidd.Measurement.PixelFootprint.Col*sidd.Display.get_pixel_size()
            memory_limit = int(numpy.floor(im_seg_limit/row_memory_size))
            row_limit.append(min(value, memory_limit))
        self._row_limit = tuple(row_limit)

    def _create_security_tags(self):
        def class_priority(cls1, cls2):
            p1 = _class_priority[cls1]
            p2 = _class_priority[cls2]
            if p1 >= p2:
                return cls1
            return cls2

        # determine the highest priority clas string from all the sidds & sicds
        clas_collection = [extract_clas(sidd) for sidd in self.sidd_meta]
        if self.sicd_meta is not None:
            clas_collection.extend([extract_clas_sicd(sicd) for sicd in self.sicd_meta])
        clas = reduce(class_priority, clas_collection)
        # try to determine clsy from all sidds
        clsy_collection = list(set(extract_clsy(sidd) for sidd in self.sidd_meta))
        clsy = clsy_collection[0] if len(clsy_collection) else None
        # populate the attribute
        self._security_tags = NITFSecurityTags(CLAS=clas, CLSY=clsy)

    def _get_iid2(self, index: int) -> str:
        """
        Get the IID2 for the sidd at `index`.

        Parameters
        ----------
        index : int

        Returns
        -------
        str
        """

        sidd = self.sidd_meta[index]

        iid2 = sidd.NITF.get('IID2', None)
        if iid2 is None:
            iid2 = sidd.NITF.get('FTITLE', None)
        if iid2 is None:
            iid2 = sidd.NITF.get('SUGGESTED_NAME', None)
        if iid2 is None:
            iid2 = 'SIDD: Unknown'
        return iid2

    def _get_ftitle(self, index: int = 0) -> str:
        sidd = self.sidd_meta[index]
        ftitle = sidd.NITF.get('FTITLE', None)
        if ftitle is None:
            ftitle = sidd.NITF.get('SUGGESTED_NAME', None)
        if ftitle is None:
            ftitle = 'SIDD: Unknown'
        return ftitle

    def _get_fdt(self, index: int) -> Optional[str]:
        sidd = self.sidd_meta[index]
        if sidd.ExploitationFeatures.Collections[0].Information.CollectionDateTime is not None:
            the_time = sidd.ExploitationFeatures.Collections[0].Information.CollectionDateTime.astype('datetime64[s]')
            return re.sub(r'[^0-9]', '', str(the_time))
        else:
            return None

    def _get_ostaid(self, index: int = 0) -> str:
        sidd = self.sidd_meta[index]
        ostaid = sidd.NITF.get('OSTAID', 'Unknown')
        return ostaid

    def _get_isorce(self, index: int = 0) -> str:
        sidd = self.sidd_meta[index]
        isorce = sidd.NITF.get('ISORCE', sidd.ExploitationFeatures.Collections[0].Information.SensorName)
        if isorce is None:
            isorce = 'Unknown'
        return isorce

    def _get_icp(self, sidd_index: int) -> Optional[numpy.ndarray]:
        """
        Get the Image corner point array, if possible.

        Parameters
        ----------
        sidd_index : int

        Returns
        -------
        None|numpy.ndarray
        """

        sidd = self.sidd_meta[sidd_index]  # type: Union[SIDDType3, SIDDType2, SIDDType1]
        if isinstance(sidd, SIDDType3) and sidd.GeoData is not None and sidd.GeoData.ImageCorners is not None:
            return sidd.GeoData.ImageCorners.get_array(dtype=numpy.dtype('float64'))
        elif isinstance(sidd, SIDDType2) and sidd.GeoData is not None and sidd.GeoData.ImageCorners is not None:
            return sidd.GeoData.ImageCorners.get_array(dtype=numpy.dtype('float64'))
        elif isinstance(sidd, SIDDType1) and sidd.GeographicAndTarget is not None and \
                sidd.GeographicAndTarget.GeographicCoverage is not None and \
                sidd.GeographicAndTarget.GeographicCoverage.Footprint is not None:
            return sidd.GeographicAndTarget.GeographicCoverage.Footprint.get_array(dtype=numpy.dtype('float64'))
        return None

    def _create_header(self) -> NITFHeader:
        """
        Create the main NITF header.

        Returns
        -------
        NITFHeader
        """

        # NB: CLEVEL and FL will be corrected...
        return NITFHeader(
            Security=self._security_tags, CLEVEL=3, OSTAID=self._get_ostaid(),
            FDT=self._get_fdt(0), FTITLE=self._get_ftitle(), FL=0)

    def _create_image_segment_for_sidd(
            self,
            sidd_index: int,
            starting_index: int) -> Tuple[List[ImageSubheaderManager], Tuple[int, ...], Tuple[Tuple[int, ...], ...]]:

        image_managers = []
        sidd = self.sidd_meta[sidd_index]

        if isinstance(sidd, (SIDDType2, SIDDType3)) and sidd.Compression is not None:
            raise ValueError('Compression not currently supported.')

        basic_args = {
            'ICAT': 'SAR',
            'IC': 'NC',
            'IID2': self._get_iid2(sidd_index),
            'ISORCE': self._get_isorce(sidd_index),
            'IDATIM': self._get_fdt(sidd_index)
        }

        if sidd.Display.PixelType == 'MONO8I':
            basic_args['PVTYPE'] = 'INT'
            basic_args['NBPP'] = 8
            basic_args['ABPP'] = 8
            basic_args['IREP'] = 'MONO'
            basic_args['IMODE'] = 'B'
            irepband = ('M', )
        elif sidd.Display.PixelType == 'MONO16I':
            basic_args['PVTYPE'] = 'INT'
            basic_args['NBPP'] = 16
            basic_args['ABPP'] = 16
            basic_args['IREP'] = 'MONO'
            basic_args['IMODE'] = 'B'
            irepband = ('M', )
        elif sidd.Display.PixelType == 'RGB24I':
            basic_args['PVTYPE'] = 'INT'
            basic_args['NBPP'] = 8
            basic_args['ABPP'] = 8
            basic_args['IREP'] = 'RGB'
            basic_args['IMODE'] = 'P'
            irepband = ('R', 'G', 'B')
        else:
            raise ValueError('Unsupported PixelType {}'.format(sidd.Display.PixelType))

        rows = sidd.Measurement.PixelFootprint.Row
        cols = sidd.Measurement.PixelFootprint.Col
        icp = self._get_icp(sidd_index)
        image_segment_limits = default_image_segmentation(rows, cols, self._row_limit[sidd_index])

        image_segment_indices = []
        total_image_count = starting_index
        for i, entry in enumerate(image_segment_limits):
            image_segment_indices.append(total_image_count)
            total_image_count += 1
            if i == 0:
                iloc = '0000000000'
            else:
                prev_lims = image_segment_limits[i-1]
                prev_rows = prev_lims[1] - prev_lims[0]
                iloc = '{0:05d}00000'.format(prev_rows)

            this_rows = entry[1]-entry[0]
            this_cols = entry[3]-entry[2]
            subhead = ImageSegmentHeader(
                IID1='SIDD{0:03d}{1:03d}'.format(sidd_index+1, i+1),
                NROWS=this_rows,
                NCOLS=this_cols,
                IGEOLO=interpolate_corner_points_string(numpy.array(entry, dtype=numpy.int64), rows, cols, icp),
                NPPBH=0 if this_cols > 8192 else this_cols,
                NPPBV=0 if this_rows > 8192 else this_rows,
                IDLVL=total_image_count,
                IALVL=0 if i == 0 else (total_image_count - 1),
                ILOC=iloc,
                Bands=ImageBands(values=[ImageBand(ISUBCAT='', IREPBAND=entry) for entry in irepband]),
                Security=self._sidd_security_tags[sidd_index],
                **basic_args)
            image_managers.append(ImageSubheaderManager(subhead))
        return image_managers, tuple(image_segment_indices), image_segment_limits

    def _create_image_segments(self) -> Tuple[Tuple[ImageSubheaderManager, ...], Tuple[Tuple[int, ...], ...], Tuple[Tuple[Tuple[int, ...], ...]]]:
        image_managers = []
        image_segment_collection = []
        image_segment_coordinates = []
        starting_index = 0
        for i in range(len(self.sidd_meta)):
            t_managers, t_indices, t_limit = self._create_image_segment_for_sidd(i, starting_index)
            image_managers.extend(t_managers)
            image_segment_collection.append(t_indices)
            image_segment_coordinates.append(t_limit)
            starting_index = t_indices[-1] + 1
        return tuple(image_managers), tuple(image_segment_collection), tuple(image_segment_coordinates)

    def _create_des_segment_for_sidd(self, sidd_index: int) -> DESSubheaderManager:
        sidd = self.sidd_meta[sidd_index]
        uh_args = sidd.get_des_details()

        try:
            desshdt = str(sidd.ProductCreation.ProcessorInformation.ProcessingDateTime.astype('datetime64[s]'))
        except AttributeError:
            desshdt = str(numpy.datetime64('now'))
        if desshdt[-1] != 'Z':
            desshdt += 'Z'
        uh_args['DESSHDT'] = desshdt

        desshlpg = ''
        icp = self._get_icp(sidd_index)
        if icp is not None:
            temp = []
            for entry in icp:
                temp.append('{0:0=+12.8f}{1:0=+13.8f}'.format(entry[0], entry[1]))
            temp.append(temp[0])
            desshlpg = ''.join(temp)
        uh_args['DESSHLPG'] = desshlpg
        subhead = DataExtensionHeader(
            Security=self._sidd_security_tags[sidd_index],
            UserHeader=XMLDESSubheader(**uh_args))
        return DESSubheaderManager(
            subhead, sidd.to_xml_bytes(tag='SIDD'))

    def _create_sidd_des_segments(self) -> List[DESSubheaderManager]:
        return [self._create_des_segment_for_sidd(index) for index in range(len(self.sidd_meta))]

    def _create_des_segment_for_sicd(self, sicd_index: int) -> DESSubheaderManager:
        sicd = self.sicd_meta[sicd_index]
        uh_args = sicd.get_des_details()

        desshdt = str(sicd.ImageCreation.DateTime.astype('datetime64[s]'))
        if desshdt[-1] != 'Z':
            desshdt += 'Z'
        uh_args['DESSHDT'] = desshdt

        desshlpg = ''
        if sicd.GeoData is not None and sicd.GeoData.ImageCorners is not None:
            # noinspection PyTypeChecker
            icp = sicd.GeoData.ImageCorners.get_array(dtype=numpy.float64)
            temp = []
            for entry in icp:
                temp.append('{0:0=+12.8f}{1:0=+13.8f}'.format(entry[0], entry[1]))
            temp.append(temp[0])
            desshlpg = ''.join(temp)
        uh_args['DESSHLPG'] = desshlpg

        subhead = DataExtensionHeader(
            Security=self._sicd_security_tags[sicd_index],
            UserHeader=XMLDESSubheader(**uh_args))
        return DESSubheaderManager(
            subhead, sicd.to_xml_bytes(tag='SICD', urn=uh_args['DESSHTN']))

    def _create_sicd_des_segments(self) -> List[DESSubheaderManager]:
        if self.sicd_meta is None:
            return []
        return [self._create_des_segment_for_sicd(index) for index in range(len(self.sicd_meta))]

    def _create_des_segments(
            self,
            additional_des: Optional[Sequence[DESSubheaderManager]]) -> Tuple[DESSubheaderManager, ...]:

        if additional_des is not None:
            des_managers = list(additional_des)
        else:
            des_managers = []
        des_managers.extend(self._create_sidd_des_segments())
        des_managers.extend(self._create_sicd_des_segments())
        return tuple(des_managers)


class SIDDWriter(NITFWriter):
    """
    Writer class for a SIDD file - a NITF file following certain rules.

    **Changed in version 1.3.0** to reflect NITFWriter changes.
    """

    def __init__(
            self,
            file_object: Union[str, BinaryIO],
            sidd_meta: Optional[Union[SIDDType3, SIDDType2, SIDDType1, Sequence[SIDDType3],
                                Sequence[SIDDType2], Sequence[SIDDType1]]] = None,
            sicd_meta: Optional[Union[SICDType, Sequence[SICDType]]] = None,
            sidd_writing_details: Optional[SIDDWritingDetails] = None,
            check_existence: bool = True):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
        sidd_meta : None|SIDDType3|SIDDType2|SIDDType1|Sequence[SIDDType3]|Sequence[SIDDType2]|Sequence[SIDDType1]
        sicd_meta : None|SICDType|Sequence[SICDType]
        sidd_writing_details : None|SIDDWritingDetails
        check_existence : bool
            Should we check if the given file already exists?
        """

        if sidd_meta is None and sidd_writing_details is None:
            raise ValueError('One of sidd_meta or sidd_writing_details must be provided.')
        if sidd_writing_details is None:
            sidd_writing_details = SIDDWritingDetails(sidd_meta, sicd_meta=sicd_meta)
        NITFWriter.__init__(
            self, file_object, sidd_writing_details, check_existence=check_existence)

    @property
    def nitf_writing_details(self) -> SIDDWritingDetails:
        """
        SIDDWritingDetails: The SIDD/NITF subheader details.
        """

        # noinspection PyTypeChecker
        return self._nitf_writing_details

    @nitf_writing_details.setter
    def nitf_writing_details(self, value):
        if self._nitf_writing_details is not None:
            raise ValueError('nitf_writing_details is read-only')
        if not isinstance(value, SIDDWritingDetails):
            raise TypeError('nitf_writing_details must be of type {}'.format(SIDDWritingDetails))
        self._nitf_writing_details = value




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class GRD(TREElement):
    def __init__(self, value):
        super(GRD, self).__init__()
        self.add_field('ZVL', 's', 10, value)
        self.add_field('BAD', 's', 10, value)
        self.add_field('LOD', 's', 12, value)
        self.add_field('LAD', 's', 12, value)
        self.add_field('LSO', 's', 11, value)
        self.add_field('PSO', 's', 11, value)


class GRDPSBType(TREElement):
    def __init__(self, value):
        super(GRDPSBType, self).__init__()
        self.add_field('NUM_GRDS', 'd', 2, value)
        self.add_loop('GRDs', self.NUM_GRDS, GRD, value)


class GRDPSB(TREExtension):
    _tag_value = 'GRDPSB'
    _data_type = GRDPSBType



"""
The SIDDType 1.0 definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Union
from collections import OrderedDict
from copy import deepcopy

from sarpy.io.xml.base import Serializable, parse_xml_from_file, parse_xml_from_string
from sarpy.io.xml.descriptors import SerializableDescriptor
from .ProductCreation import ProductCreationType
from .Display import ProductDisplayType
from .GeographicAndTarget import GeographicAndTargetType
from .Measurement import MeasurementType
from .ExploitationFeatures import ExploitationFeaturesType
from sarpy.io.product.sidd2_elements.DownstreamReprocessing import DownstreamReprocessingType
from sarpy.io.product.sidd2_elements.ProductProcessing import ProductProcessingType
from sarpy.io.product.sidd2_elements.Annotations import AnnotationsType
from sarpy.io.product.sidd2_elements.blocks import ErrorStatisticsType, RadiometricType
from sarpy.geometry import point_projection
from sarpy.io.product.sidd_schema import get_specification_identifier, \
    get_urn_details, validate_xml_ns


DEFAULT_STRICT = False
logger = logging.getLogger(__name__)

############
# namespace validate and definition of required entries in the namespace dictionary
_SIDD_SPECIFICATION_IDENTIFIER = get_specification_identifier()

_SIDD_URN = 'urn:SIDD:1.0.0'
_sidd_details = get_urn_details(_SIDD_URN)
_SIDD_SPECIFICATION_VERSION = _sidd_details['version']
_SIDD_SPECIFICATION_DATE = _sidd_details['date']
_ISM_URN = _sidd_details['ism_urn']
_SFA_URN = _sidd_details['sfa_urn']
_SICOMMON_URN = _sidd_details['sicommon_urn']


##########
# The SIDD object

class SIDDType(Serializable):
    """
    The root element of the SIDD 1.0 document.
    """

    _fields = (
        'ProductCreation', 'Display', 'GeographicAndTarget', 'Measurement', 'ExploitationFeatures',
        'DownstreamReprocessing', 'ErrorStatistics', 'Radiometric', 'ProductProcessing', 'Annotations')
    _required = (
        'ProductCreation', 'Display', 'GeographicAndTarget', 'Measurement', 'ExploitationFeatures')
    # Descriptor
    ProductCreation = SerializableDescriptor(
        'ProductCreation', ProductCreationType, _required, strict=DEFAULT_STRICT,
        docstring='Information related to processor, classification, '
                  'and product type.')  # type: ProductCreationType
    Display = SerializableDescriptor(
        'Display', ProductDisplayType, _required, strict=DEFAULT_STRICT,
        docstring='Contains information on the parameters needed to display the product in '
                  'an exploitation tool.')  # type: ProductDisplayType
    GeographicAndTarget = SerializableDescriptor(
        'GeographicAndTarget', GeographicAndTargetType, _required, strict=DEFAULT_STRICT,
        docstring='Contains generic and extensible targeting and geographic region '
                  'information.')  # type: GeographicAndTargetType
    Measurement = SerializableDescriptor(
        'Measurement', MeasurementType, _required, strict=DEFAULT_STRICT,
        docstring='Contains the metadata necessary for performing '
                  'measurements.')  # type: MeasurementType
    ExploitationFeatures = SerializableDescriptor(
        'ExploitationFeatures', ExploitationFeaturesType, _required, strict=DEFAULT_STRICT,
        docstring='Computed metadata regarding the input collections and '
                  'final product.')  # type: ExploitationFeaturesType
    DownstreamReprocessing = SerializableDescriptor(
        'DownstreamReprocessing', DownstreamReprocessingType, _required, strict=DEFAULT_STRICT,
        docstring='Metadata describing any downstream processing of the '
                  'product.')  # type: Union[None, DownstreamReprocessingType]
    ErrorStatistics = SerializableDescriptor(
        'ErrorStatistics', ErrorStatisticsType, _required, strict=DEFAULT_STRICT,
        docstring='Error statistics passed through from the SICD '
                  'metadata.')  # type: Union[None, ErrorStatisticsType]
    Radiometric = SerializableDescriptor(
        'Radiometric', RadiometricType, _required, strict=DEFAULT_STRICT,
        docstring='Radiometric information about the product.')  # type: Union[None, RadiometricType]
    ProductProcessing = SerializableDescriptor(
        'ProductProcessing', ProductProcessingType, _required, strict=DEFAULT_STRICT,
        docstring='Contains metadata related to algorithms used during '
                  'product generation.')  # type: ProductProcessingType
    Annotations = SerializableDescriptor(
        'Annotations', AnnotationsType, _required, strict=DEFAULT_STRICT,
        docstring='List of annotations for the imagery.')  # type: AnnotationsType

    def __init__(self, ProductCreation=None, Display=None, GeographicAndTarget=None,
                 Measurement=None, ExploitationFeatures=None, DownstreamReprocessing=None,
                 ErrorStatistics=None, Radiometric=None, ProductProcessing=None,
                 Annotations=None, **kwargs):
        """

        Parameters
        ----------
        ProductCreation : ProductCreationType
        Display : ProductDisplayType
        GeographicAndTarget : GeographicAndTargetType
        Measurement : MeasurementType
        ExploitationFeatures : ExploitationFeaturesType
        DownstreamReprocessing : None|DownstreamReprocessingType
        ErrorStatistics : None|ErrorStatisticsType
        Radiometric : None|RadiometricType
        ProductProcessing : None|ProductProcessingType
        Annotations : None|AnnotationsType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']

        nitf = kwargs.get('_NITF', {})
        if not isinstance(nitf, dict):
            raise TypeError('Provided NITF options are required to be in dictionary form.')
        self._NITF = nitf

        self._coa_projection = None
        self.ProductCreation = ProductCreation
        self.Display = Display
        self.GeographicAndTarget = GeographicAndTarget
        self.Measurement = Measurement
        self.ExploitationFeatures = ExploitationFeatures
        self.DownstreamReprocessing = DownstreamReprocessing
        self.ErrorStatistics = ErrorStatistics
        self.Radiometric = Radiometric
        self.ProductProcessing = ProductProcessing
        self.Annotations = Annotations
        super(SIDDType, self).__init__(**kwargs)

    @property
    def coa_projection(self):
        """
        The COA Projection object, if previously defined through using :func:`define_coa_projection`.

        Returns
        -------
        None|sarpy.geometry.point_projection.COAProjection
        """

        return self._coa_projection

    @property
    def NITF(self):
        """
        Optional dictionary of NITF header information, pertains only to subsequent
        SIDD file writing.

        Returns
        -------
        Dict
        """

        return self._NITF

    def can_project_coordinates(self):
        """
        Determines whether the necessary elements are populated to permit projection
        between image and physical coordinates. If False, then the (first discovered)
        reason why not will be logged at error level.

        Returns
        -------
        bool
        """

        if self._coa_projection is not None:
            return True

        if self.Measurement.ProjectionType != 'PlaneProjection':
            logger.error(
                'Formulating a projection is only supported for PlaneProjection, '
                'got {}.'.format(self.Measurement.ProjectionType))
            return False
        return True

    def define_coa_projection(self, delta_arp=None, delta_varp=None, range_bias=None,
                              adj_params_frame='ECF', override=True):
        """
        Define the COAProjection object.

        Parameters
        ----------
        delta_arp : None|numpy.ndarray|list|tuple
            ARP position adjustable parameter (ECF, m).  Defaults to 0 in each coordinate.
        delta_varp : None|numpy.ndarray|list|tuple
            VARP position adjustable parameter (ECF, m/s).  Defaults to 0 in each coordinate.
        range_bias : float|int
            Range bias adjustable parameter (m), defaults to 0.
        adj_params_frame : str
            One of ['ECF', 'RIC_ECF', 'RIC_ECI'], specifying the coordinate frame used for
            expressing `delta_arp` and `delta_varp` parameters.
        override : bool
            should we redefine, if it is previously defined?

        Returns
        -------
        None
        """

        if not self.can_project_coordinates():
            logger.error('The COAProjection object cannot be defined.')
            return

        if self._coa_projection is not None and not override:
            return

        self._coa_projection = point_projection.COAProjection.from_sidd(
            self, delta_arp=delta_arp, delta_varp=delta_varp, range_bias=range_bias,
            adj_params_frame=adj_params_frame)

    def project_ground_to_image(self, coords, **kwargs):
        """
        Transforms a 3D ECF point to pixel (row/column) coordinates. This is
        implemented in accordance with the SICD Image Projections Description Document.
        **Really Scene-To-Image projection.**"

        Parameters
        ----------
        coords : numpy.ndarray|tuple|list
            ECF coordinate to map to scene coordinates, of size `N x 3`.
        kwargs
            The keyword arguments for the :func:`sarpy.geometry.point_projection.ground_to_image` method.

        Returns
        -------
        Tuple[numpy.ndarray, float, int]
            * `image_points` - the determined image point array, of size `N x 2`. Following
              the SICD convention, he upper-left pixel is [0, 0].
            * `delta_gpn` - residual ground plane displacement (m).
            * `iterations` - the number of iterations performed.

        See Also
        --------
        sarpy.geometry.point_projection.ground_to_image
        """

        if 'use_structure_coa' not in kwargs:
            kwargs['use_structure_coa'] = True
        return point_projection.ground_to_image(coords, self, **kwargs)

    def project_ground_to_image_geo(self, coords, ordering='latlong', **kwargs):
        """
        Transforms a 3D Lat/Lon/HAE point to pixel (row/column) coordinates. This is
        implemented in accordance with the SICD Image Projections Description Document.
        **Really Scene-To-Image projection.**"

        Parameters
        ----------
        coords : numpy.ndarray|tuple|list
            ECF coordinate to map to scene coordinates, of size `N x 3`.
        ordering : str
            If 'longlat', then the input is `[longitude, latitude, hae]`.
            Otherwise, the input is `[latitude, longitude, hae]`. Passed through
            to :func:`sarpy.geometry.geocoords.geodetic_to_ecf`.
        kwargs
            The keyword arguments for the :func:`sarpy.geometry.point_projection.ground_to_image_geo` method.

        Returns
        -------
        Tuple[numpy.ndarray, float, int]
            * `image_points` - the determined image point array, of size `N x 2`. Following
              the SICD convention, he upper-left pixel is [0, 0].
            * `delta_gpn` - residual ground plane displacement (m).
            * `iterations` - the number of iterations performed.

        See Also
        --------
        sarpy.geometry.point_projection.ground_to_image_geo
        """

        if 'use_structure_coa' not in kwargs:
            kwargs['use_structure_coa'] = True
        return point_projection.ground_to_image_geo(coords, self, ordering=ordering, **kwargs)

    def project_image_to_ground(self, im_points, projection_type='HAE', **kwargs):
        """
        Transforms image coordinates to ground plane ECF coordinate via the algorithm(s)
        described in SICD Image Projections document.

        Parameters
        ----------
        im_points : numpy.ndarray|list|tuple
            the image coordinate array
        projection_type : str
            One of `['PLANE', 'HAE', 'DEM']`. Type `DEM` is a work in progress.
        kwargs
            The keyword arguments for the :func:`sarpy.geometry.point_projection.image_to_ground` method.

        Returns
        -------
        numpy.ndarray
            Ground Plane Point (in ECF coordinates) corresponding to the input image coordinates.

        See Also
        --------
        sarpy.geometry.point_projection.image_to_ground
        """

        if 'use_structure_coa' not in kwargs:
            kwargs['use_structure_coa'] = True
        return point_projection.image_to_ground(
            im_points, self, projection_type=projection_type, **kwargs)

    def project_image_to_ground_geo(self, im_points, ordering='latlong', projection_type='HAE', **kwargs):
        """
        Transforms image coordinates to ground plane WGS-84 coordinate via the algorithm(s)
        described in SICD Image Projections document.

        Parameters
        ----------
        im_points : numpy.ndarray|list|tuple
            the image coordinate array
        projection_type : str
            One of `['PLANE', 'HAE', 'DEM']`. Type `DEM` is a work in progress.
        ordering : str
            Determines whether return is ordered as `[lat, long, hae]` or `[long, lat, hae]`.
            Passed through to :func:`sarpy.geometry.geocoords.ecf_to_geodetic`.
        kwargs
            The keyword arguments for the :func:`sarpy.geometry.point_projection.image_to_ground_geo` method.

        Returns
        -------
        numpy.ndarray
            Ground Plane Point (in ECF coordinates) corresponding to the input image coordinates.
        See Also
        --------
        sarpy.geometry.point_projection.image_to_ground_geo
        """

        if 'use_structure_coa' not in kwargs:
            kwargs['use_structure_coa'] = True
        return point_projection.image_to_ground_geo(
            im_points, self, ordering=ordering, projection_type=projection_type, **kwargs)

    @staticmethod
    def get_xmlns_collection():
        """
        Gets the correct SIDD 1.0 dictionary of xml namespace details.

        Returns
        -------
        dict
        """

        return OrderedDict([
            ('xmlns', _SIDD_URN), ('xmlns:sicommon', _SICOMMON_URN),
            ('xmlns:sfa', _SFA_URN), ('xmlns:ism', _ISM_URN)])

    @staticmethod
    def get_des_details():
        """
        Gets the correct SIDD 1.0 DES subheader details.

        Returns
        -------
        dict
        """

        return OrderedDict([
            ('DESSHSI', _SIDD_SPECIFICATION_IDENTIFIER),
            ('DESSHSV', _SIDD_SPECIFICATION_VERSION),
            ('DESSHSD', _SIDD_SPECIFICATION_DATE),
            ('DESSHTN', _SIDD_URN)])

    @classmethod
    def from_node(cls, node, xml_ns, ns_key='default', kwargs=None):
        if ns_key is None:
            raise ValueError('ns_key must be defined.')
        if ns_key not in xml_ns:
            raise ValueError('ns_key {} is not in the xml namespace'.format(ns_key))

        valid_ns = validate_xml_ns(xml_ns, ns_key)
        if not xml_ns[ns_key].startswith('urn:SIDD:1.'):
            raise ValueError('Cannot use urn {} for SIDD version 1.0'.format(xml_ns[ns_key]))
        if not valid_ns:
            logger.warning(
                'SIDD namespace validation failed,\n\t'
                'which may lead to subsequent deserialization failures')
        return super(SIDDType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_xml_bytes(self, urn=None, tag='SIDD', check_validity=False, strict=DEFAULT_STRICT):
        if urn is None:
            urn = self.get_xmlns_collection()
        return super(SIDDType, self).to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict)

    def to_xml_string(self, urn=None, tag='SIDD', check_validity=False, strict=DEFAULT_STRICT):
        return self.to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict).decode('utf-8')

    def copy(self):
        """
        Provides a deep copy.

        Returns
        -------
        SIDDType
        """

        out = super(SIDDType, self).copy()
        out._NITF = deepcopy(self._NITF)
        return out

    @classmethod
    def from_xml_file(cls, file_path):
        """
        Construct the sidd object from a stand-alone xml file path.

        Parameters
        ----------
        file_path : str

        Returns
        -------
        SIDDType
        """

        root_node, xml_ns = parse_xml_from_file(file_path)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)

    @classmethod
    def from_xml_string(cls, xml_string):
        """
        Construct the sidd object from an xml string.

        Parameters
        ----------
        xml_string : str|bytes

        Returns
        -------
        SIDDType
        """

        root_node, xml_ns = parse_xml_from_string(xml_string)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)
